#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

from __future__ import unicode_literals

import unittest

from psamm.datasource import sbml, native, entry
from psamm.datasource.reaction import parse_reaction
from psamm.reaction import Reaction, Compound, Direction

from decimal import Decimal
from fractions import Fraction
from six import BytesIO, itervalues


class TestSBMLDatabaseL1V2(unittest.TestCase):
    """Test parsing of a simple level 1 version 2 SBML file"""

    def setUp(self):
        self.doc = BytesIO('''<?xml version="1.0" encoding="UTF-8"?>
<sbml xmlns="http://www.sbml.org/sbml/level1"
      xmlns:html="http://www.w3.org/1999/xhtml"
      level="1" version="2">
 <model name="Test model">
  <listOfCompartments>
   <compartment name="boundary"/>
   <compartment name="cell"/>
  </listOfCompartments>
  <listOfSpecies>
   <species name="Glucose" compartment="cell" initialAmount="1"
            charge="0"/>
   <species name="Glucose_6_P" compartment="cell" initialAmount="1"
            charge="-2"/>
   <species name="H2O" compartment="cell" initialAmount="1" charge="0"/>
   <species name="Phosphate" compartment="cell" initialAmount="1"
            boundaryCondition="false"/>
   <species name="Biomass" compartment="boundary" initialAmount="1"
            boundaryCondition="true"/>
  </listOfSpecies>
  <listOfReactions>
   <reaction name="G6Pase" reversible="true">
    <listOfReactants>
     <speciesReference species="Glucose" stoichiometry="2"/>
     <speciesReference species="Phosphate" stoichiometry="2"/>
    </listOfReactants>
    <listOfProducts>
     <speciesReference species="H2O" stoichiometry="2"/>
     <speciesReference species="Glucose_6_P" stoichiometry="2"/>
    </listOfProducts>
    <notes>
     <html:p>Glucose 6-phosphatase</html:p>
    </notes>
   </reaction>
   <reaction name="Biomass" reversible="false">
    <listOfReactants>
     <speciesReference species="Glucose_6_P" stoichiometry="56"
                       denominator="100"/>
     <speciesReference species="Glucose" stoichiometry="88"
                       denominator="100"/>
    </listOfReactants>
    <listOfProducts>
     <speciesReference species="Biomass"/>
    </listOfProducts>
   </reaction>
  </listOfReactions>
 </model>
</sbml>'''.encode('utf-8'))

    def test_model_name(self):
        reader = sbml.SBMLReader(self.doc)
        self.assertEqual(reader.name, 'Test model')

    def test_compartment_exists(self):
        reader = sbml.SBMLReader(self.doc, ignore_boundary=False)
        compartments = {entry.id: entry for entry in reader.compartments}
        self.assertEqual(len(compartments), 2)
        self.assertEqual(compartments['cell'].id, 'cell')
        self.assertEqual(compartments['cell'].name, 'cell')
        self.assertEqual(compartments['boundary'].id, 'boundary')
        self.assertEqual(compartments['boundary'].name, 'boundary')

    def test_compartment_exists_with_ignore_boundary(self):
        reader = sbml.SBMLReader(self.doc, ignore_boundary=True)
        compartments = {entry.id: entry for entry in reader.compartments}
        self.assertEqual(len(compartments), 1)
        self.assertEqual(compartments['cell'].id, 'cell')
        self.assertEqual(compartments['cell'].name, 'cell')

    def test_compounds_exist(self):
        reader = sbml.SBMLReader(self.doc, ignore_boundary=False)
        species = {entry.id: entry for entry in reader.species}
        self.assertEqual(len(species), 5)

        self.assertEqual(species['Glucose'].id, 'Glucose')
        self.assertEqual(species['Glucose'].name, 'Glucose')
        self.assertEqual(species['Glucose'].compartment, 'cell')
        self.assertFalse(species['Glucose'].boundary)
        self.assertEqual(species['Glucose'].charge, 0)

        self.assertEqual(species['Glucose_6_P'].id, 'Glucose_6_P')
        self.assertEqual(species['Glucose_6_P'].name, 'Glucose_6_P')
        self.assertEqual(species['Glucose_6_P'].compartment, 'cell')
        self.assertFalse(species['Glucose_6_P'].boundary)
        self.assertEqual(species['Glucose_6_P'].charge, -2)

        self.assertEqual(species['H2O'].id, 'H2O')
        self.assertEqual(species['H2O'].name, 'H2O')
        self.assertEqual(species['H2O'].compartment, 'cell')
        self.assertFalse(species['H2O'].boundary)
        self.assertEqual(species['H2O'].charge, 0)

        self.assertFalse(species['Phosphate'].boundary)
        self.assertTrue(species['Biomass'].boundary)

    def test_g6pase_reaction_exists(self):
        reader = sbml.SBMLReader(self.doc)
        reaction = reader.get_reaction('G6Pase')
        self.assertTrue(reaction.reversible)

        # Compare equation of reaction
        actual_equation = Reaction(Direction.Both,
                                   [(Compound('Glucose', 'cell'), 2),
                                    (Compound('Phosphate', 'cell'), 2)],
                                   [(Compound('H2O', 'cell'), 2),
                                    (Compound('Glucose_6_P', 'cell'), 2)])
        self.assertEqual(reaction.equation, actual_equation)

    def test_biomass_reaction_exists(self):
        reader = sbml.SBMLReader(self.doc, ignore_boundary=False)
        reaction = reader.get_reaction('Biomass')
        self.assertFalse(reaction.reversible)

        # Compare equation of reaction
        actual_equation = Reaction(Direction.Forward, [
            (Compound('Glucose_6_P', 'cell'), Fraction(56, 100)),
            (Compound('Glucose', 'cell'), Fraction(88, 100))
        ], [
            (Compound('Biomass', 'boundary'), 1)
        ])
        self.assertEqual(reaction.equation, actual_equation)

    def test_reaction_xml_notes(self):
        reader = sbml.SBMLReader(self.doc, ignore_boundary=False)
        reaction = reader.get_reaction('G6Pase')
        notes = reaction.xml_notes

        notes_tags = list(notes)
        self.assertEqual(len(notes_tags), 1)
        self.assertEqual(notes_tags[0].tag, '{http://www.w3.org/1999/xhtml}p')
        self.assertEqual(notes_tags[0].text, 'Glucose 6-phosphatase')

    def test_objective_not_present(self):
        reader = sbml.SBMLReader(self.doc)
        objectives = list(reader.objectives)
        self.assertEqual(len(objectives), 0)
        self.assertIsNone(reader.get_active_objective())

    def test_flux_bounds_not_present(self):
        reader = sbml.SBMLReader(self.doc)
        flux_bounds = list(reader.flux_bounds)
        self.assertEqual(len(flux_bounds), 0)

    def test_create_and_convert_model(self):
        reader = sbml.SBMLReader(self.doc)
        model = reader.create_model()
        sbml.convert_sbml_model(model)

        self.assertEqual(
            {entry.id for entry in model.compounds},
            {'Glucose', 'Glucose_6_P', 'H2O', 'Phosphate'})
        self.assertEqual(
            {entry.id for entry in model.reactions},
            {'G6Pase', 'Biomass'})
        self.assertEqual(
            {entry.id for entry in model.compartments},
            {'cell'})

        self.assertEqual(set(model.model), {'Biomass', 'G6Pase'})


class TestSBMLDatabaseL2V5(unittest.TestCase):
    """Test parsing of a simple level 2 version 5 SBML file"""

    def setUp(self):
        self.doc = BytesIO('''<?xml version="1.0" encoding="UTF-8"?>
<sbml xmlns="http://www.sbml.org/sbml/level2/version5"
      level="2" version="5">
 <model id="test_model" name="Test model">
  <listOfCompartments>
   <compartment id="C_c" name="cell"/>
   <compartment id="C_b" name="boundary"/>
  </listOfCompartments>
  <listOfSpecies>
   <species id="M_Glucose_LPAREN_c_RPAREN_" name="Glucose" compartment="C_c"
            charge="0">
    <notes>
     <body xmlns="http://www.w3.org/1999/xhtml">
      <p>FORMULA: C6H12O6</p>
      <p>Charge: "0"</p>
      <p>Additional notes..</p>
      <p>KEGG ID: C00031</p>
      <p>Custom: 123</p>
     </body>
    </notes>
   </species>
   <species id="M_Glucose_6_DASH_P_LPAREN_c_RPAREN_" name="Glucose-6-P"
            compartment="C_c" charge="-2"/>
   <species id="M_H2O_LPAREN_c_RPAREN_" name="H2O" compartment="C_c"
            charge="0"/>
   <species id="M_Phosphate_LPAREN_c_RPAREN_" name="Phosphate"
            compartment="C_c" boundaryCondition="false"/>
   <species id="M_Biomass" name="Biomass" compartment="C_b"
            boundaryCondition="true"/>
  </listOfSpecies>
  <listOfReactions>
   <reaction id="R_G6Pase" reversible="true">
    <listOfReactants>
     <speciesReference species="M_Glucose_LPAREN_c_RPAREN_" stoichiometry="2"/>
     <speciesReference species="M_Phosphate_LPAREN_c_RPAREN_"
                       stoichiometry="2"/>
    </listOfReactants>
    <listOfProducts>
     <speciesReference species="M_H2O_LPAREN_c_RPAREN_" stoichiometry="2"/>
     <speciesReference species="M_Glucose_6_DASH_P_LPAREN_c_RPAREN_"
                       stoichiometry="2"/>
    </listOfProducts>
    <notes>
     <body xmlns="http://www.w3.org/1999/xhtml">
      <p> Authors: Jane Doe ; John Doe </p>
      <p>CONFIDENCE: 3</p>
      <p>EC NUMBER: 3.1.3.9</p>
      <p>gene association :   b0822 </p>
      <p>SUBSYSTEM: "Glycolysis / Gluconeogenesis"</p>
      <p>Additional notes...</p>
     </body>
    </notes>
   </reaction>
   <reaction id="R_Biomass" reversible="false">
    <listOfReactants>
     <speciesReference species="M_Glucose_6_DASH_P_LPAREN_c_RPAREN_"
                       stoichiometry="0.56"/>
     <speciesReference species="M_Glucose_LPAREN_c_RPAREN_"
                       stoichiometry="0.88"/>
    </listOfReactants>
    <listOfProducts>
     <speciesReference species="M_Biomass"/>
    </listOfProducts>
    <kineticLaw>
     <listOfParameters>
      <parameter id="LOWER_BOUND" value="0"/>
      <parameter id="UPPER_BOUND" value="1000"/>
      <parameter id="SOME_CUSTOM_PARAMETER" value="123.4"/>
      <parameter id="OBJECTIVE_COEFFICIENT" value="1"/>
     </listOfParameters>
    </kineticLaw>
   </reaction>
  </listOfReactions>
 </model>
</sbml>'''.encode('utf-8'))

    def test_model_name(self):
        reader = sbml.SBMLReader(self.doc)
        self.assertEqual(reader.id, 'test_model')
        self.assertEqual(reader.name, 'Test model')

    def test_compartment_exists(self):
        reader = sbml.SBMLReader(self.doc, ignore_boundary=False)
        compartments = {entry.id: entry for entry in reader.compartments}
        self.assertEqual(len(compartments), 2)
        self.assertEqual(compartments['C_c'].id, 'C_c')
        self.assertEqual(compartments['C_c'].name, 'cell')
        self.assertEqual(compartments['C_b'].id, 'C_b')
        self.assertEqual(compartments['C_b'].name, 'boundary')

    def test_compartment_exists_with_ignore_boundary(self):
        reader = sbml.SBMLReader(self.doc, ignore_boundary=True)
        compartments = {entry.id: entry for entry in reader.compartments}
        self.assertEqual(len(compartments), 1)
        self.assertEqual(compartments['C_c'].id, 'C_c')
        self.assertEqual(compartments['C_c'].name, 'cell')

    def test_compounds_exist(self):
        reader = sbml.SBMLReader(self.doc, ignore_boundary=False)
        species = {entry.id: entry for entry in reader.species}
        self.assertEqual(len(species), 5)

        gluc_species = species['M_Glucose_LPAREN_c_RPAREN_']
        self.assertEqual(gluc_species.id, 'M_Glucose_LPAREN_c_RPAREN_')
        self.assertEqual(gluc_species.name, 'Glucose')
        self.assertEqual(gluc_species.compartment, 'C_c')
        self.assertFalse(gluc_species.boundary)
        self.assertEqual(gluc_species.charge, 0)

        g6p_species = species['M_Glucose_6_DASH_P_LPAREN_c_RPAREN_']
        self.assertEqual(g6p_species.id, 'M_Glucose_6_DASH_P_LPAREN_c_RPAREN_')
        self.assertEqual(g6p_species.name, 'Glucose-6-P')
        self.assertEqual(g6p_species.compartment, 'C_c')
        self.assertFalse(g6p_species.boundary)
        self.assertEqual(g6p_species.charge, -2)

        h2o_species = species['M_H2O_LPAREN_c_RPAREN_']
        self.assertEqual(h2o_species.id, 'M_H2O_LPAREN_c_RPAREN_')
        self.assertEqual(h2o_species.name, 'H2O')
        self.assertEqual(h2o_species.compartment, 'C_c')
        self.assertFalse(h2o_species.boundary)
        self.assertEqual(h2o_species.charge, 0)

        self.assertFalse(species['M_Phosphate_LPAREN_c_RPAREN_'].boundary)
        self.assertTrue(species['M_Biomass'].boundary)

    def test_glucose_parse_notes(self):
        reader = sbml.SBMLReader(self.doc)
        species = reader.get_species('M_Glucose_LPAREN_c_RPAREN_')
        notes_dict = sbml.parse_xhtml_species_notes(species)
        self.assertEqual(notes_dict, {
            'formula': 'C6H12O6',
            'kegg': 'C00031',
            'charge': 0
        })

    def test_g6pase_reaction_exists(self):
        reader = sbml.SBMLReader(self.doc)
        reaction = reader.get_reaction('R_G6Pase')
        self.assertTrue(reaction.reversible)

        # Compare equation of reaction
        actual_equation = Reaction(Direction.Both, [
            (Compound('M_Glucose_LPAREN_c_RPAREN_', 'C_c'), 2),
            (Compound('M_Phosphate_LPAREN_c_RPAREN_', 'C_c'), 2)
        ], [
            (Compound('M_H2O_LPAREN_c_RPAREN_', 'C_c'), 2),
            (Compound('M_Glucose_6_DASH_P_LPAREN_c_RPAREN_', 'C_c'), 2)
        ])
        self.assertEqual(reaction.equation, actual_equation)

    def test_g6pase_parse_notes(self):
        reader = sbml.SBMLReader(self.doc)
        reaction = reader.get_reaction('R_G6Pase')
        notes = sbml.parse_xhtml_reaction_notes(reaction)
        self.assertEqual(notes, {
            'subsystem': 'Glycolysis / Gluconeogenesis',
            'genes': 'b0822',
            'ec': '3.1.3.9',
            'confidence': 3,
            'authors': ['Jane Doe', 'John Doe']
        })

    def test_parse_reaction_cobra_flux_bounds(self):
        reader = sbml.SBMLReader(self.doc)
        reaction = reader.get_reaction('R_G6Pase')
        lower, upper = sbml.parse_flux_bounds(reaction)
        self.assertIsNone(lower)
        self.assertIsNone(upper)

        reaction = reader.get_reaction('R_Biomass')
        lower, upper = sbml.parse_flux_bounds(reaction)
        self.assertEqual(lower, 0)
        self.assertEqual(upper, 1000)

    def test_parse_reaction_cobra_objective(self):
        reader = sbml.SBMLReader(self.doc)
        reaction = reader.get_reaction('R_G6Pase')
        coeff = sbml.parse_objective_coefficient(reaction)
        self.assertIsNone(coeff)

        reaction = reader.get_reaction('R_Biomass')
        coeff = sbml.parse_objective_coefficient(reaction)
        self.assertEqual(coeff, 1)

    def test_biomass_reaction_exists(self):
        reader = sbml.SBMLReader(self.doc, ignore_boundary=False)
        reaction = reader.get_reaction('R_Biomass')
        self.assertFalse(reaction.reversible)

        # Compare equation of reaction
        actual_equation = Reaction(Direction.Forward, [
            (Compound('M_Glucose_6_DASH_P_LPAREN_c_RPAREN_', 'C_c'),
             Decimal('0.56')),
            (Compound('M_Glucose_LPAREN_c_RPAREN_', 'C_c'), Decimal('0.88'))
        ], [
            (Compound('M_Biomass', 'C_b'), 1)
        ])
        self.assertEqual(reaction.equation, actual_equation)

    def test_objective_not_present(self):
        reader = sbml.SBMLReader(self.doc)
        objectives = list(reader.objectives)
        self.assertEqual(len(objectives), 0)
        self.assertIsNone(reader.get_active_objective())

    def test_flux_bounds_not_present(self):
        reader = sbml.SBMLReader(self.doc)
        flux_bounds = list(reader.flux_bounds)
        self.assertEqual(len(flux_bounds), 0)

    def test_create_and_convert_model(self):
        reader = sbml.SBMLReader(self.doc)
        model = reader.create_model()
        sbml.convert_sbml_model(model)

        self.assertEqual(
            {entry.id for entry in model.compounds},
            {'Glucose(c)', 'Glucose_6-P(c)', 'H2O(c)', 'Phosphate(c)'})
        self.assertEqual(
            {entry.id for entry in model.reactions},
            {'G6Pase', 'Biomass'})
        self.assertEqual(
            {entry.id for entry in model.compartments},
            {'c'})

        self.assertEqual(model.limits['Biomass'], ('Biomass', 0, 1000))
        self.assertEqual(model.biomass_reaction, 'Biomass')
        self.assertEqual(set(model.model), {'Biomass', 'G6Pase'})


class TestSBMLDatabaseL3V1(unittest.TestCase):
    """Test parsing of a simple level 3 version 1 SBML file"""

    def setUp(self):
        self.doc = BytesIO('''<?xml version="1.0" encoding="UTF-8"?>
<sbml xmlns="http://www.sbml.org/sbml/level3/version1/core"
      xmlns:html="http://www.w3.org/1999/xhtml"
      level="3" version="1">
 <model id="test_model" name="Test model">
  <listOfCompartments>
   <compartment id="C_c" name="cell" constant="true"/>
   <compartment id="C_b" name="boundary" constant="true"/>
  </listOfCompartments>
  <listOfSpecies>
   <species id="M_Glucose_LPAREN_c_RPAREN_" name="Glucose" compartment="C_c"
            constant="false" boundaryCondition="false"
            hasOnlySubstanceUnits="false"/>
   <species id="M_Glucose_6_DASH_P_LPAREN_c_RPAREN_" name="Glucose-6-P"
            compartment="C_c" constant="false" boundaryCondition="false"
            hasOnlySubstanceUnits="false"/>
   <species id="M_H2O_LPAREN_c_RPAREN_" name="H2O" compartment="C_c"
            constant="false" boundaryCondition="false"
            hasOnlySubstanceUnits="false"/>
   <species id="M_Phosphate_LPAREN_c_RPAREN_" name="Phosphate"
            compartment="C_c" constant="false" boundaryCondition="false"
            hasOnlySubstanceUnits="false"/>
   <species id="M_Biomass" name="Biomass" compartment="C_b" constant="false"
            boundaryCondition="true" hasOnlySubstanceUnits="false"/>
  </listOfSpecies>
  <listOfReactions>
   <reaction id="R_G6Pase" reversible="true" fast="false">
    <listOfReactants>
     <speciesReference species="M_Glucose_LPAREN_c_RPAREN_" stoichiometry="2"
                       constant="true"/>
     <speciesReference species="M_Phosphate_LPAREN_c_RPAREN_" stoichiometry="2"
                       constant="true"/>
    </listOfReactants>
    <listOfProducts>
     <speciesReference species="M_H2O_LPAREN_c_RPAREN_" stoichiometry="2"
                       constant="true"/>
     <speciesReference species="M_Glucose_6_DASH_P_LPAREN_c_RPAREN_"
                       stoichiometry="2" constant="true"/>
    </listOfProducts>
    <notes>
     <html:p>Glucose 6-phosphatase</html:p>
    </notes>
   </reaction>
   <reaction id="R_Biomass" reversible="false" fast="false">
    <listOfReactants>
     <speciesReference species="M_Glucose_6_DASH_P_LPAREN_c_RPAREN_"
                       stoichiometry="0.56" constant="true"/>
     <speciesReference species="M_Glucose_LPAREN_c_RPAREN_"
                       stoichiometry="0.88" constant="true"/>
    </listOfReactants>
    <listOfProducts>
     <speciesReference species="M_Biomass" stoichiometry="1" constant="true"/>
    </listOfProducts>
   </reaction>
  </listOfReactions>
 </model>
</sbml>'''.encode('utf-8'))

    def test_model_name(self):
        reader = sbml.SBMLReader(self.doc)
        self.assertEqual(reader.id, 'test_model')
        self.assertEqual(reader.name, 'Test model')

    def test_compartment_exists(self):
        reader = sbml.SBMLReader(self.doc, ignore_boundary=False)
        compartments = {entry.id: entry for entry in reader.compartments}
        self.assertEqual(len(compartments), 2)
        self.assertEqual(compartments['C_c'].id, 'C_c')
        self.assertEqual(compartments['C_c'].name, 'cell')
        self.assertEqual(compartments['C_b'].id, 'C_b')
        self.assertEqual(compartments['C_b'].name, 'boundary')

    def test_compartment_exists_with_ignore_boundary(self):
        reader = sbml.SBMLReader(self.doc, ignore_boundary=True)
        compartments = {entry.id: entry for entry in reader.compartments}
        self.assertEqual(len(compartments), 1)
        self.assertEqual(compartments['C_c'].id, 'C_c')
        self.assertEqual(compartments['C_c'].name, 'cell')

    def test_compounds_exist(self):
        reader = sbml.SBMLReader(self.doc, ignore_boundary=False)
        species = {entry.id: entry for entry in reader.species}
        self.assertEqual(len(species), 5)

        gluc_species = species['M_Glucose_LPAREN_c_RPAREN_']
        self.assertEqual(gluc_species.id, 'M_Glucose_LPAREN_c_RPAREN_')
        self.assertEqual(gluc_species.name, 'Glucose')
        self.assertEqual(gluc_species.compartment, 'C_c')
        self.assertFalse(gluc_species.boundary)

        g6p_species = species['M_Glucose_6_DASH_P_LPAREN_c_RPAREN_']
        self.assertEqual(g6p_species.id, 'M_Glucose_6_DASH_P_LPAREN_c_RPAREN_')
        self.assertEqual(g6p_species.name, 'Glucose-6-P')
        self.assertEqual(g6p_species.compartment, 'C_c')
        self.assertFalse(g6p_species.boundary)

        h2o_species = species['M_H2O_LPAREN_c_RPAREN_']
        self.assertEqual(h2o_species.id, 'M_H2O_LPAREN_c_RPAREN_')
        self.assertEqual(h2o_species.name, 'H2O')
        self.assertEqual(h2o_species.compartment, 'C_c')
        self.assertFalse(h2o_species.boundary)

        self.assertFalse(species['M_Phosphate_LPAREN_c_RPAREN_'].boundary)
        self.assertTrue(species['M_Biomass'].boundary)

    def test_g6pase_reaction_exists(self):
        reader = sbml.SBMLReader(self.doc)
        reaction = reader.get_reaction('R_G6Pase')
        self.assertTrue(reaction.reversible)

        # Compare equation of reaction
        actual_equation = Reaction(Direction.Both, [
            (Compound('M_Glucose_LPAREN_c_RPAREN_', 'C_c'), 2),
            (Compound('M_Phosphate_LPAREN_c_RPAREN_', 'C_c'), 2)
        ], [
            (Compound('M_H2O_LPAREN_c_RPAREN_', 'C_c'), 2),
            (Compound('M_Glucose_6_DASH_P_LPAREN_c_RPAREN_', 'C_c'), 2)
        ])
        self.assertEqual(reaction.equation, actual_equation)

    def test_biomass_reaction_exists(self):
        reader = sbml.SBMLReader(self.doc, ignore_boundary=False)
        reaction = reader.get_reaction('R_Biomass')
        self.assertFalse(reaction.reversible)

        # Compare equation of reaction
        actual_equation = Reaction(Direction.Forward, [
            (Compound('M_Glucose_6_DASH_P_LPAREN_c_RPAREN_', 'C_c'),
             Decimal('0.56')),
            (Compound('M_Glucose_LPAREN_c_RPAREN_', 'C_c'), Decimal('0.88'))
        ], [
            (Compound('M_Biomass', 'C_b'), 1)
        ])
        self.assertEqual(reaction.equation, actual_equation)

    def test_reaction_xml_notes(self):
        reader = sbml.SBMLReader(self.doc)
        reaction = reader.get_reaction('R_G6Pase')
        notes = reaction.xml_notes

        notes_tags = list(notes)
        self.assertEqual(len(notes_tags), 1)
        self.assertEqual(notes_tags[0].tag, '{http://www.w3.org/1999/xhtml}p')
        self.assertEqual(notes_tags[0].text, 'Glucose 6-phosphatase')

    def test_objective_not_present(self):
        reader = sbml.SBMLReader(self.doc)
        objectives = list(reader.objectives)
        self.assertEqual(len(objectives), 0)
        self.assertIsNone(reader.get_active_objective())

    def test_flux_bounds_not_present(self):
        reader = sbml.SBMLReader(self.doc)
        flux_bounds = list(reader.flux_bounds)
        self.assertEqual(len(flux_bounds), 0)

    def test_create_and_convert_model(self):
        reader = sbml.SBMLReader(self.doc)
        model = reader.create_model()
        sbml.convert_sbml_model(model)

        self.assertEqual(
            {entry.id for entry in model.compounds},
            {'Glucose(c)', 'Glucose_6-P(c)', 'H2O(c)', 'Phosphate(c)'})
        self.assertEqual(
            {entry.id for entry in model.reactions},
            {'G6Pase', 'Biomass'})
        self.assertEqual(
            {entry.id for entry in model.compartments},
            {'c'})

        self.assertEqual(set(model.model), {'Biomass', 'G6Pase'})


class TestSBMLDatabaseL3V1WithFBCV1(unittest.TestCase):
    """Test parsing of a level 3 version 1 SBML file with FBC version 1"""

    def setUp(self):
        self.doc = BytesIO('''<?xml version="1.0" encoding="UTF-8"?>
<sbml xmlns="http://www.sbml.org/sbml/level3/version1/core"
      xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version1"
      xmlns:html="http://www.w3.org/1999/xhtml"
      level="3" version="1"
      fbc:required="false">
 <model id="test_model" name="Test model">
  <listOfCompartments>
   <compartment id="C_c" name="cell" constant="true"/>
   <compartment id="C_b" name="boundary" constant="true"/>
  </listOfCompartments>
  <listOfSpecies>
   <species id="M_Glucose_LPAREN_c_RPAREN_" name="Glucose" compartment="C_c"
            constant="false" boundaryCondition="false"
            hasOnlySubstanceUnits="false" fbc:charge="0"
            fbc:chemicalFormula="C6H12O6"/>
   <species id="M_Glucose_6_DASH_P_LPAREN_c_RPAREN_" name="Glucose-6-P"
            compartment="C_c" constant="false" boundaryCondition="false"
            hasOnlySubstanceUnits="false" fbc:charge="-2"
            fbc:chemicalFormula="C6H11O9P"/>
   <species id="M_H2O_LPAREN_c_RPAREN_" name="H2O" compartment="C_c"
            constant="false" boundaryCondition="false"
            hasOnlySubstanceUnits="false" fbc:charge="0"
            fbc:chemicalFormula="H2O"/>
   <species id="M_Phosphate_LPAREN_c_RPAREN_" name="Phosphate"
            compartment="C_c" constant="false" boundaryCondition="false"
            hasOnlySubstanceUnits="false" fbc:charge="-2"
            fbc:chemicalFormula="HO4P"/>
   <species id="M_Biomass" name="Biomass" compartment="C_b" constant="false"
            boundaryCondition="true" hasOnlySubstanceUnits="false"/>
  </listOfSpecies>
  <listOfReactions>
   <reaction id="R_G6Pase" reversible="true" fast="false">
    <listOfReactants>
     <speciesReference species="M_Glucose_LPAREN_c_RPAREN_" stoichiometry="2"
                       constant="true"/>
     <speciesReference species="M_Phosphate_LPAREN_c_RPAREN_" stoichiometry="2"
                       constant="true"/>
    </listOfReactants>
    <listOfProducts>
     <speciesReference species="M_H2O_LPAREN_c_RPAREN_" stoichiometry="2"
                       constant="true"/>
     <speciesReference species="M_Glucose_6_DASH_P_LPAREN_c_RPAREN_"
                       stoichiometry="2" constant="true"/>
    </listOfProducts>
   </reaction>
   <reaction id="R_Biomass" reversible="false" fast="false">
    <listOfReactants>
     <speciesReference species="M_Glucose_6_DASH_P_LPAREN_c_RPAREN_"
                       stoichiometry="0.56" constant="true"/>
     <speciesReference species="M_Glucose_LPAREN_c_RPAREN_"
                       stoichiometry="0.88" constant="true"/>
    </listOfReactants>
    <listOfProducts>
     <speciesReference species="M_Biomass" stoichiometry="1" constant="true"/>
    </listOfProducts>
   </reaction>
  </listOfReactions>
  <fbc:listOfObjectives fbc:activeObjective="obj1">
   <fbc:objective fbc:id="obj1" fbc:name="Objective 1" fbc:type="maximize">
    <fbc:listOfFluxObjectives>
     <fbc:fluxObjective fbc:reaction="R_Biomass" fbc:coefficient="1"/>
    </fbc:listOfFluxObjectives>
   </fbc:objective>
  </fbc:listOfObjectives>
  <fbc:listOfFluxBounds>
   <fbc:fluxBound fbc:reaction="R_G6Pase" fbc:operation="greaterEqual"
                  fbc:value="-10"/>
   <fbc:fluxBound fbc:reaction="R_G6Pase" fbc:operation="lessEqual"
                  fbc:value="1000"/>
   <fbc:fluxBound fbc:reaction="R_Biomass" fbc:operation="greaterEqual"
                  fbc:value="0"/>
   <fbc:fluxBound fbc:reaction="R_Biomass" fbc:operation="lessEqual"
                  fbc:value="1000"/>
  </fbc:listOfFluxBounds>
 </model>
</sbml>'''.encode('utf-8'))

    def test_model_name(self):
        reader = sbml.SBMLReader(self.doc)
        self.assertEqual(reader.id, 'test_model')
        self.assertEqual(reader.name, 'Test model')

    def test_compartment_exists(self):
        reader = sbml.SBMLReader(self.doc, ignore_boundary=False)
        compartments = {entry.id: entry for entry in reader.compartments}
        self.assertEqual(len(compartments), 2)
        self.assertEqual(compartments['C_c'].id, 'C_c')
        self.assertEqual(compartments['C_c'].name, 'cell')
        self.assertEqual(compartments['C_b'].id, 'C_b')
        self.assertEqual(compartments['C_b'].name, 'boundary')

    def test_compounds_exist(self):
        reader = sbml.SBMLReader(self.doc, ignore_boundary=False)
        species = {entry.id: entry for entry in reader.species}
        self.assertEqual(len(species), 5)

        gluc_species = species['M_Glucose_LPAREN_c_RPAREN_']
        self.assertEqual(gluc_species.id, 'M_Glucose_LPAREN_c_RPAREN_')
        self.assertEqual(gluc_species.name, 'Glucose')
        self.assertEqual(gluc_species.compartment, 'C_c')
        self.assertFalse(gluc_species.boundary)
        self.assertEqual(gluc_species.formula, 'C6H12O6')
        self.assertEqual(gluc_species.charge, 0)

        g6p_species = species['M_Glucose_6_DASH_P_LPAREN_c_RPAREN_']
        self.assertEqual(g6p_species.id, 'M_Glucose_6_DASH_P_LPAREN_c_RPAREN_')
        self.assertEqual(g6p_species.name, 'Glucose-6-P')
        self.assertEqual(g6p_species.compartment, 'C_c')
        self.assertFalse(g6p_species.boundary)
        self.assertEqual(g6p_species.formula, 'C6H11O9P')
        self.assertEqual(g6p_species.charge, -2)

        h2o_species = species['M_H2O_LPAREN_c_RPAREN_']
        self.assertEqual(h2o_species.id, 'M_H2O_LPAREN_c_RPAREN_')
        self.assertEqual(h2o_species.name, 'H2O')
        self.assertEqual(h2o_species.compartment, 'C_c')
        self.assertFalse(h2o_species.boundary)
        self.assertEqual(h2o_species.formula, 'H2O')
        self.assertEqual(h2o_species.charge, 0)

        self.assertFalse(species['M_Phosphate_LPAREN_c_RPAREN_'].boundary)
        self.assertTrue(species['M_Biomass'].boundary)

        self.assertIsNone(species['M_Biomass'].formula)
        self.assertIsNone(species['M_Biomass'].charge)

    def test_g6pase_reaction_exists(self):
        reader = sbml.SBMLReader(self.doc, ignore_boundary=False)
        reaction = reader.get_reaction('R_G6Pase')
        self.assertTrue(reaction.reversible)

        # Compare equation of reaction
        actual_equation = Reaction(Direction.Both, [
            (Compound('M_Glucose_LPAREN_c_RPAREN_', 'C_c'), 2),
            (Compound('M_Phosphate_LPAREN_c_RPAREN_', 'C_c'), 2)
        ], [
            (Compound('M_H2O_LPAREN_c_RPAREN_', 'C_c'), 2),
            (Compound('M_Glucose_6_DASH_P_LPAREN_c_RPAREN_', 'C_c'), 2)
        ])
        self.assertEqual(reaction.equation, actual_equation)

        self.assertEqual(reaction.properties['lower_flux'], -10)
        self.assertEqual(reaction.properties['upper_flux'], 1000)

    def test_biomass_reaction_exists(self):
        reader = sbml.SBMLReader(self.doc, ignore_boundary=False)
        reaction = reader.get_reaction('R_Biomass')
        self.assertFalse(reaction.reversible)

        # Compare equation of reaction
        actual_equation = Reaction(Direction.Forward, [
            (Compound('M_Glucose_6_DASH_P_LPAREN_c_RPAREN_', 'C_c'),
             Decimal('0.56')),
            (Compound('M_Glucose_LPAREN_c_RPAREN_', 'C_c'), Decimal('0.88'))
        ], [
            (Compound('M_Biomass', 'C_b'), 1)
        ])
        self.assertEqual(reaction.equation, actual_equation)

        self.assertEqual(reaction.properties['lower_flux'], 0)
        self.assertEqual(reaction.properties['upper_flux'], 1000)

    def test_objective_exists(self):
        reader = sbml.SBMLReader(self.doc)
        objectives = {entry.id: entry for entry in reader.objectives}
        self.assertEqual(len(objectives), 1)

        objective = objectives['obj1']
        self.assertEqual(objective.name, 'Objective 1')
        self.assertEqual(objective.type, 'maximize')
        self.assertEqual(dict(objective.reactions), {'R_Biomass': 1})

    def test_active_objective(self):
        reader = sbml.SBMLReader(self.doc)
        objectives = {entry.id: entry for entry in reader.objectives}
        self.assertEqual(reader.get_active_objective(), objectives['obj1'])

    def test_flux_bounds_exists(self):
        reader = sbml.SBMLReader(self.doc)
        flux_bounds = list(reader.flux_bounds)
        self.assertEqual(len(flux_bounds), 4)

        biomass_bounds = set(
            (b.operation, b.value) for b in flux_bounds
            if b.reaction == 'R_Biomass')
        self.assertEqual(biomass_bounds, {
            ('greaterEqual', 0),
            ('lessEqual', 1000)})

        g6pase_bounds = set(
            (b.operation, b.value) for b in flux_bounds
            if b.reaction == 'R_G6Pase')
        self.assertEqual(g6pase_bounds, {
            ('greaterEqual', -10),
            ('lessEqual', 1000)})

    def test_create_and_convert_model(self):
        reader = sbml.SBMLReader(self.doc)
        model = reader.create_model()
        sbml.convert_sbml_model(model)

        self.assertEqual(
            {entry.id for entry in model.compounds},
            {'Glucose(c)', 'Glucose_6-P(c)', 'H2O(c)', 'Phosphate(c)'})
        self.assertEqual(
            {entry.id for entry in model.reactions},
            {'G6Pase', 'Biomass'})
        self.assertEqual(
            {entry.id for entry in model.compartments},
            {'c'})

        self.assertEqual(model.limits['Biomass'], ('Biomass', 0, 1000))
        self.assertEqual(model.limits['G6Pase'], ('G6Pase', -10, 1000))
        self.assertEqual(set(model.model), {'Biomass', 'G6Pase'})
        self.assertEqual(model.biomass_reaction, 'Biomass')


class TestSBMLDatabaseL3V1WithFBCV2(unittest.TestCase):
    """Test parsing of a level 3 version 1 SBML file with FBC version 2"""

    def setUp(self):
        self.doc = BytesIO('''<?xml version="1.0" encoding="UTF-8"?>
<sbml xmlns="http://www.sbml.org/sbml/level3/version1/core"
      xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2"
      xmlns:html="http://www.w3.org/1999/xhtml"
      level="3" version="1"
      fbc:required="false">
 <model id="test_model" name="Test model">
  <notes>
   <body xmlns="http://www.w3.org/1999/xhtml">
    <p>This is model information intended to be seen by humans.</p>
   </body>
  </notes>
  <listOfCompartments>
   <compartment id="C_c" name="cell" constant="true"/>
   <compartment id="C_b" name="boundary" constant="true"/>
  </listOfCompartments>
  <listOfSpecies>
   <species id="M_Glucose_LPAREN_c_RPAREN_"
            metaid="meta_M_Glucose_LPAREN_c_RPAREN_" name="Glucose"
            compartment="C_c" constant="false" boundaryCondition="false"
            hasOnlySubstanceUnits="false" fbc:charge="0"
            fbc:chemicalFormula="C6H12O6">
    <notes>
     <body xmlns="http://www.w3.org/1999/xhtml">
      <p>This is compound information intended to be seen by humans.</p>
     </body>
    </notes>
   </species>
   <species id="M_Glucose_6_DASH_P_LPAREN_c_RPAREN_" name="Glucose-6-P"
            compartment="C_c" constant="false" boundaryCondition="false"
            hasOnlySubstanceUnits="false" fbc:charge="-2"
            fbc:chemicalFormula="C6H11O9P"/>
   <species id="M_H2O_LPAREN_c_RPAREN_" name="H2O" compartment="C_c"
            constant="false" boundaryCondition="false"
            hasOnlySubstanceUnits="false" fbc:charge="0"
            fbc:chemicalFormula="H2O"/>
   <species id="M_Phosphate_LPAREN_c_RPAREN_" name="Phosphate"
            compartment="C_c" constant="false" boundaryCondition="false"
            hasOnlySubstanceUnits="false" fbc:charge="-2"
            fbc:chemicalFormula="HO4P"/>
   <species id="M_Biomass" name="Biomass" compartment="C_b" constant="false"
            boundaryCondition="true" hasOnlySubstanceUnits="false"/>
  </listOfSpecies>
  <listOfParameters>
   <parameter constant="true" id="P_lower_G6Pase" value="-10"/>
   <parameter constant="true" id="P_lower_zero" value="0"/>
   <parameter constant="true" id="P_upper_default" value="1000"/>
  </listOfParameters>
  <listOfReactions>
   <reaction id="R_G6Pase" metaid="meta_R_G6Pase" reversible="true"
             fast="false" fbc:lowerFluxBound="P_lower_G6Pase"
             fbc:upperFluxBound="P_upper_default">
    <listOfReactants>
     <speciesReference species="M_Glucose_LPAREN_c_RPAREN_" stoichiometry="2"
                       constant="true"/>
     <speciesReference species="M_Phosphate_LPAREN_c_RPAREN_" stoichiometry="2"
                       constant="true"/>
    </listOfReactants>
    <listOfProducts>
     <speciesReference species="M_H2O_LPAREN_c_RPAREN_" stoichiometry="2"
                       constant="true"/>
     <speciesReference species="M_Glucose_6_DASH_P_LPAREN_c_RPAREN_"
                       stoichiometry="2" constant="true"/>
    </listOfProducts>
    <notes>
     <body xmlns="http://www.w3.org/1999/xhtml">
      <p>This is reaction information intended to be seen by humans.</p>
     </body>
    </notes>
    <annotation>
     <RDF xmlns="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
          xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
      <Description about="#meta_R_G6Pase">
       <bqbiol:isVersionOf>
        <Bag>
         <li resource="http://identifiers.org/ec-code/3.1.3.9"/>
        </Bag>
       </bqbiol:isVersionOf>
      </Description>
     </RDF>
    </annotation>
   </reaction>
   <reaction id="R_Biomass" reversible="false" fast="false"
             fbc:lowerFluxBound="P_lower_zero"
             fbc:upperFluxBound="P_upper_default">
    <listOfReactants>
     <speciesReference species="M_Glucose_6_DASH_P_LPAREN_c_RPAREN_"
                       stoichiometry="0.56" constant="true"/>
     <speciesReference species="M_Glucose_LPAREN_c_RPAREN_"
                       stoichiometry="0.88" constant="true"/>
    </listOfReactants>
    <listOfProducts>
     <speciesReference species="M_Biomass" stoichiometry="1" constant="true"/>
    </listOfProducts>
   </reaction>
  </listOfReactions>
  <fbc:listOfObjectives fbc:activeObjective="obj1">
   <fbc:objective fbc:id="obj1" fbc:name="Objective 1" fbc:type="maximize">
    <fbc:listOfFluxObjectives>
     <fbc:fluxObjective fbc:reaction="R_Biomass" fbc:coefficient="1"/>
    </fbc:listOfFluxObjectives>
   </fbc:objective>
  </fbc:listOfObjectives>
 </model>
</sbml>'''.encode('utf-8'))

    def test_model_name(self):
        reader = sbml.SBMLReader(self.doc)
        self.assertEqual(reader.id, 'test_model')
        self.assertEqual(reader.name, 'Test model')

    def test_compartment_exists(self):
        reader = sbml.SBMLReader(self.doc, ignore_boundary=False)
        compartments = {entry.id: entry for entry in reader.compartments}
        self.assertEqual(len(compartments), 2)
        self.assertEqual(compartments['C_c'].id, 'C_c')
        self.assertEqual(compartments['C_c'].name, 'cell')
        self.assertEqual(compartments['C_b'].id, 'C_b')
        self.assertEqual(compartments['C_b'].name, 'boundary')

    def test_compounds_exist(self):
        reader = sbml.SBMLReader(self.doc, ignore_boundary=False)
        species = {entry.id: entry for entry in reader.species}
        self.assertEqual(len(species), 5)

        gluc_species = species['M_Glucose_LPAREN_c_RPAREN_']
        self.assertEqual(gluc_species.id, 'M_Glucose_LPAREN_c_RPAREN_')
        self.assertEqual(gluc_species.name, 'Glucose')
        self.assertEqual(gluc_species.compartment, 'C_c')
        self.assertFalse(gluc_species.boundary)
        self.assertEqual(gluc_species.formula, 'C6H12O6')
        self.assertEqual(gluc_species.charge, 0)
        self.assertIsNotNone(gluc_species.xml_notes)

        g6p_species = species['M_Glucose_6_DASH_P_LPAREN_c_RPAREN_']
        self.assertEqual(g6p_species.id, 'M_Glucose_6_DASH_P_LPAREN_c_RPAREN_')
        self.assertEqual(g6p_species.name, 'Glucose-6-P')
        self.assertEqual(g6p_species.compartment, 'C_c')
        self.assertFalse(g6p_species.boundary)
        self.assertEqual(g6p_species.formula, 'C6H11O9P')
        self.assertEqual(g6p_species.charge, -2)
        self.assertIsNone(g6p_species.xml_notes)

        h2o_species = species['M_H2O_LPAREN_c_RPAREN_']
        self.assertEqual(h2o_species.id, 'M_H2O_LPAREN_c_RPAREN_')
        self.assertEqual(h2o_species.name, 'H2O')
        self.assertEqual(h2o_species.compartment, 'C_c')
        self.assertFalse(h2o_species.boundary)
        self.assertEqual(h2o_species.formula, 'H2O')
        self.assertEqual(h2o_species.charge, 0)

        self.assertFalse(species['M_Phosphate_LPAREN_c_RPAREN_'].boundary)
        self.assertTrue(species['M_Biomass'].boundary)

        self.assertIsNone(species['M_Biomass'].formula)
        self.assertIsNone(species['M_Biomass'].charge)

    def test_g6pase_reaction_exists(self):
        reader = sbml.SBMLReader(self.doc)
        reaction = reader.get_reaction('R_G6Pase')
        self.assertTrue(reaction.reversible)

        # Compare equation of reaction
        actual_equation = Reaction(Direction.Both, [
            (Compound('M_Glucose_LPAREN_c_RPAREN_', 'C_c'), 2),
            (Compound('M_Phosphate_LPAREN_c_RPAREN_', 'C_c'), 2)
        ], [
            (Compound('M_H2O_LPAREN_c_RPAREN_', 'C_c'), 2),
            (Compound('M_Glucose_6_DASH_P_LPAREN_c_RPAREN_', 'C_c'), 2)
        ])
        self.assertEqual(reaction.equation, actual_equation)

        self.assertEqual(reaction.properties['lower_flux'], -10)
        self.assertEqual(reaction.properties['upper_flux'], 1000)

        self.assertIsNotNone(reaction.xml_notes)
        self.assertIsNotNone(reaction.xml_annotation)

    def test_biomass_reaction_exists(self):
        reader = sbml.SBMLReader(self.doc, ignore_boundary=False)
        reaction = reader.get_reaction('R_Biomass')
        self.assertFalse(reaction.reversible)

        # Compare equation of reaction
        actual_equation = Reaction(Direction.Forward, [
            (Compound('M_Glucose_6_DASH_P_LPAREN_c_RPAREN_', 'C_c'),
             Decimal('0.56')),
            (Compound('M_Glucose_LPAREN_c_RPAREN_', 'C_c'), Decimal('0.88'))
        ], [
            (Compound('M_Biomass', 'C_b'), 1)
        ])
        self.assertEqual(reaction.equation, actual_equation)

        self.assertEqual(reaction.properties['lower_flux'], 0)
        self.assertEqual(reaction.properties['upper_flux'], 1000)

        self.assertIsNone(reaction.xml_notes)
        self.assertIsNone(reaction.xml_annotation)

    def test_objective_exists(self):
        reader = sbml.SBMLReader(self.doc)
        objectives = {entry.id: entry for entry in reader.objectives}
        self.assertEqual(len(objectives), 1)

        objective = objectives['obj1']
        self.assertEqual(objective.name, 'Objective 1')
        self.assertEqual(objective.type, 'maximize')
        self.assertEqual(dict(objective.reactions), {'R_Biomass': 1})

    def test_active_objective(self):
        reader = sbml.SBMLReader(self.doc)
        objectives = {entry.id: entry for entry in reader.objectives}
        self.assertEqual(reader.get_active_objective(),
                         objectives['obj1'])

    def test_flux_bounds_not_present(self):
        reader = sbml.SBMLReader(self.doc)
        flux_bounds = list(reader.flux_bounds)
        self.assertEqual(len(flux_bounds), 0)

    def test_create_and_convert_model(self):
        reader = sbml.SBMLReader(self.doc)
        model = reader.create_model()
        sbml.convert_sbml_model(model)

        self.assertEqual(
            {entry.id for entry in model.compounds},
            {'Glucose(c)', 'Glucose_6-P(c)', 'H2O(c)', 'Phosphate(c)'})
        self.assertEqual(
            {entry.id for entry in model.reactions},
            {'G6Pase', 'Biomass'})
        self.assertEqual(
            {entry.id for entry in model.compartments},
            {'c'})

        self.assertEqual(model.limits['Biomass'], ('Biomass', 0, 1000))
        self.assertEqual(model.limits['G6Pase'], ('G6Pase', -10, 1000))
        self.assertEqual(set(model.model), {'Biomass', 'G6Pase'})
        self.assertEqual(model.biomass_reaction, 'Biomass')


class TestModelExtracellularCompartment(unittest.TestCase):
    def setUp(self):
        self.model = native.NativeModel()
        self.model.reactions.update([
            entry.DictReactionEntry({
                'id': 'EX_g6p',
                'equation': parse_reaction('g6p[e] <=>')
            }),
            entry.DictReactionEntry({
                'id': 'EX_glc-D',
                'equation': parse_reaction('glc-D[e] <=>')
            }),
            entry.DictReactionEntry({
                'id': 'SINK_glc-D',
                'equation': parse_reaction('glc-D[c] <=>')
            }),
            entry.DictReactionEntry({
                'id': 'rxn_1',
                'equation': parse_reaction('g6p[c] <=> glc-D[c]')
            }),
            entry.DictReactionEntry({
                'id': 'TP_glc-D',
                'equation': parse_reaction('glc-D[c] <=> glc-D[e]')
            })
        ])

    def test_detect_model(self):
        """Test that the extracellular compartment is detected.

        Since the 'e' compartment occurs more often in exchange reactions,
        this compartment should be detected.
        """
        extracellular = sbml.detect_extracellular_compartment(self.model)
        self.assertEqual(extracellular, 'e')

    def test_convert_model(self):
        self.model.extracellular_compartment = 'e'
        sbml.convert_exchange_to_compounds(self.model)

        self.assertEqual(
            {entry.id for entry in self.model.reactions},
            {'rxn_1', 'TP_glc-D', 'SINK_glc-D'})
        self.assertEqual(
            set(itervalues(self.model.exchange)),
            {
                (Compound('g6p', 'e'), 'EX_g6p', None, None),
                (Compound('glc-D', 'e'), 'EX_glc-D', None, None)
            })


class TestMergeEquivalentCompounds(unittest.TestCase):
    def test_merge(self):
        model = native.NativeModel()
        model.compounds.update([
            entry.DictCompoundEntry({
                'id': 'g6p_c',
                'name': 'G6P'
            }),
            entry.DictCompoundEntry({
                'id': 'g6p_e',
                'name': 'G6P'
            })
        ])
        model.reactions.update([
            entry.DictReactionEntry({
                'id': 'TP_g6p',
                'equation': parse_reaction('g6p_c[c] <=> g6p_e[e]')
            })
        ])
        exchange_compound = Compound('g6p_e', 'e')
        model.exchange[exchange_compound] = (
            exchange_compound, 'EX_g6p_e', -10, 0)

        sbml.merge_equivalent_compounds(model)

        self.assertEqual({entry.id for entry in model.compounds}, {'g6p'})
        self.assertEqual(
            model.reactions['TP_g6p'].equation,
            parse_reaction('g6p[c] <=> g6p[e]'))

        new_exchange_compound = Compound('g6p', 'e')
        self.assertEqual(
            model.exchange[new_exchange_compound],
            (new_exchange_compound, 'EX_g6p_e', -10, 0))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2020  Jing Wang <wjingsjtu@gmail.com>

import unittest
import tempfile
import shutil
import os

from psamm import translate_id as tr_id
from psamm.datasource.native import ModelReader, NativeModel
from psamm.datasource.reaction import parse_reaction


class TestTranslateId(unittest.TestCase):
    def setUp(self):
        self._model_dir = tempfile.mkdtemp()
        with open(os.path.join(self._model_dir, 'model.yaml'), 'w') as f:
            f.write('\n'.join([
                '---',
                'biomass: rxn_1',
                'compartments:',
                '  - id: c',
                '    adjacent_to: e',
                '  - id: e',
                'reactions:',
                '  - id: rxn_1',
                '    name: rxn_1',
                '    equation: A[e] => B[c]',
                '    genes: gene1 and gene2 or gene3',
                '  - id: rxn_2',
                '    name: rxn_2',
                '    equation: B[c] => C[e]',
                '    genes: gene5 or gene6',
                '  - id: rxn_3',
                '    name: rxn_3',
                '    equation: A[e] + (6) B[c] <=> (6) C[e] + (6) D[c]',
                '    genes: gene7',
                'compounds:',
                '  - id: A',
                '    name: compound_A',
                '    formula: C6H12O6',
                '    charge: 1',
                '  - id: B',
                '    name: compound_B',
                '    formula: O2',
                '    charge: 1',
                '    kegg: C00010',
                '  - id: C',
                '    name: compound_C',
                '    formula: CO2',
                '    charge: -1',
                '  - id: D',
                '    name: compound_D',
                '    formula: H2O',
                'exchange:',
                '  - compartment: e',
                '  - compounds:',
                '    - id: A',
                '      upper: 100',
                '    - id: C',
                'limits:',
                '  - reaction: rxn_1',
                '    lower: -50',
                '  - reaction: rxn_2',
                '    upper: 100',
            ]))
        self._model = ModelReader.reader_from_path(
            os.path.join(self._model_dir, 'model.yaml')).create_model()

        with open(os.path.join(self._model_dir, 'newmodel.yaml'), 'w') as f:
            f.write('\n'.join([
                '---',
                'biomass: rxn_1a',
                'extracellular: s',
                'compartments:',
                '  - id: c',
                '    adjacent_to: s',
                '  - id: s',
                'reactions:',
                '  - id: rxn_1a',
                '    name: rxn_1',
                '    equation: A1[s] => B2[c]',
                '    genes: gene1 and gene2 or gene3',
                '  - id: rxn_2b',
                '    name: rxn_2',
                '    equation: B2[c] => A1[s]',
                '    genes: gene5 or gene6',
                '  - id: rxn_3c',
                '    name: rxn_3',
                '    equation: A1[s] + (6) B2[c] <=> (6) A1[s] + (6) D4[c]',
                '    genes: gene7',
                'compounds:',
                '  - id: A1',
                '    name: compound_A',
                '    formula: C6H12O6',
                '    charge: 1',
                '  - id: B2',
                '    name: compound_B',
                '    formula: O2',
                '    charge: 1',
                '    kegg: C00010',
                '  - id: C3',
                '    name: compound_C',
                '    formula: CO2',
                '    charge: -1',
                '  - id: D4',
                '    name: compound_D',
                '    formula: H2O',
                'exchange:',
                '  - compartment: s',
                '  - compounds:',
                '    - id: A1',
                '      upper: 100',
                '      reaction: EX_A1(s)',
                'limits:',
                '  - reaction: rxn_1a',
                '    lower: -50',
            ]))
        self._newmodel = ModelReader.reader_from_path(
            os.path.join(self._model_dir, 'newmodel.yaml')).create_model()

        with open(os.path.join(self._model_dir, 'compount_map.tsv'), 'w') as f:
            f.write('\n'.join([
                'id1\tid2\tp',
                'A\tA1\t1.0',
                'B\tB2\t0.9',
                'C\tA1\t0.001',
                'D\tD4\t0.32',
                'F\t\t',
            ]))

        with open(os.path.join(self._model_dir, 'reaction_map.tsv'), 'w') as f:
            f.write('\n'.join([
                'id1\tid2\tp',
                'rxn_1\trxn_1a\t1.0',
                'rxn_2\trxn_1a\t0.9',
                'rxn_4\t\t',
                'rxn_3\trxn_3c\t0.32',
            ]))

        with open(
                os.path.join(self._model_dir, 'compartment_map.tsv'),
                'w') as f:
            f.write('\n'.join([
                'e\ts',
            ]))

        self._compound_map = tr_id.read_mapping(
            os.path.join(self._model_dir, 'compount_map.tsv'))
        self._reaction_map = tr_id.read_mapping(
            os.path.join(self._model_dir, 'reaction_map.tsv'))
        self._compartment_map = tr_id.read_mapping(
            os.path.join(self._model_dir, 'compartment_map.tsv'))
        self._translated = tr_id.TranslatedModel(
            self._model, self._compound_map,
            self._reaction_map, self._compartment_map)

    def tearDown(self):
        shutil.rmtree(self._model_dir)

    def test_instance(self):
        self.assertIsInstance(self._translated, NativeModel)
        self.assertIsInstance(self._translated, tr_id.TranslatedModel)

    def test_translate_compound(self):
        for cpd in ['A', 'B', 'C', 'D', 'C3']:
            self.assertNotIn(cpd, self._translated.compounds)
        for cpd in ['A1', 'B2', 'D4']:
            self.assertIn(cpd, self._translated.compounds)

    def test_translate_compartment(self):
        self.assertIn('c', self._translated.compartments)
        self.assertIn('s', self._translated.compartments)
        self.assertNotIn('e', self._translated.compartments)
        self.assertSetEqual(
            self._translated._compartment_boundaries, set({('c', 's')}))
        self.assertEqual(self._translated.extracellular_compartment, 's')

    def test_translate_reaction(self):
        for rxn in ['rxn_1', 'rxn_2', 'rxn_3', 'rxn_2b']:
            self.assertNotIn(rxn, self._translated.reactions)
        for rxn in ['rxn_1a', 'rxn_3c']:
            self.assertIn(rxn, self._translated.reactions)
            self.assertEqual(self._translated.reactions[rxn].equation,
                             self._newmodel.reactions[rxn].equation)

    def test_translate_exchange(self):
        self.assertDictEqual(self._translated.exchange,
                             self._newmodel.exchange)

    def test_translate_limits(self):
        self.assertDictEqual(self._translated.limits,
                             self._newmodel.limits)

    def test_write_model(self):
        self._translated.write_model(
            os.path.join(self._model_dir, 'translated'))

    def test_no_compartment_change(self):
        model = ModelReader.reader_from_path(
            os.path.join(self._model_dir, 'model.yaml')).create_model()
        translated = tr_id.TranslatedModel(
            model,
            self._translated.cpd_mapping_id,
            self._translated.rxn_mapping_id)
        self.assertEqual(translated.reactions['rxn_1a'].equation,
                         parse_reaction('A1[e] => B2[c]'))
        self.assertEqual(translated.reactions['rxn_3c'].equation,
                         parse_reaction(
                             'A1[e] + (6) B2[c] <=> (6) A1[e] + (6) D4[c]'))
        self.assertIn('c', translated.compartments)
        self.assertNotIn('s', translated.compartments)
        self.assertIn('e', translated.compartments)
        self.assertSetEqual(
            translated._compartment_boundaries, set({('c', 'e')}))
        self.assertEqual(translated.extracellular_compartment, 'e')



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2016  Chao liu <lcddzyx@gmail.com>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

import os
import unittest
import tempfile
import shutil
import math

from psamm.formula import Formula
from psamm.datasource.native import ModelReader
from psamm import balancecheck
from psamm.datasource.reaction import parse_reaction


class TestBalanceCheckWithModel(unittest.TestCase):
    def setUp(self):
        self._model_dir = tempfile.mkdtemp()
        with open(os.path.join(self._model_dir, 'model.yaml'), 'w') as f:
            f.write('\n'.join([
                '---',
                'reactions:',
                '  - id: rxn_1',
                '    equation: A[e] => B[c]',
                '  - id: rxn_2',
                '    equation: B[c] => C[e]',
                '  - id: rxn_3',
                '    equation: A[e] + (6) B[c] <=> (6) C[e] + (6) D[c]',
                'compounds:',
                '  - id: A',
                '    formula: C6H12O6',
                '    charge: 1',
                '  - id: B',
                '    formula: O2',
                '    charge: 1',
                '  - id: C',
                '    formula: CO2',
                '    charge: -1',
                '  - id: D',
                '    formula: H2O',
            ]))
        self._model = ModelReader.reader_from_path(
            self._model_dir).create_model()

    def tearDown(self):
        shutil.rmtree(self._model_dir)

    def test_charge_balance(self):
        d = {reaction.id: value for reaction, value in
             balancecheck.charge_balance(self._model)}
        self.assertEqual(d['rxn_1'], 0)
        self.assertEqual(d['rxn_2'], -2)
        self.assertTrue(math.isnan(d['rxn_3']))

    def test_formula_balance(self):
        d = {reaction.id: value for reaction, value in
             balancecheck.formula_balance(self._model)}
        self.assertEqual(d['rxn_1'][0], Formula.parse('C6H12O6'))
        self.assertEqual(d['rxn_1'][1], Formula.parse('O2'))

        self.assertEqual(d['rxn_2'][0], Formula.parse('O2'))
        self.assertEqual(d['rxn_2'][1], Formula.parse('CO2'))

        self.assertEqual(d['rxn_3'][0], Formula.parse('C6H12O18'))
        self.assertEqual(d['rxn_3'][1], Formula.parse('C6H12O18'))


class TestBalanceCheckWithReaction(unittest.TestCase):
    def test_reaction_charge_zero_sum(self):
        reaction = parse_reaction('A[e] + (6) B[c] <=> (6) C[e] + (6) D[c]')
        compound_charge = {'A': 6, 'B': -1, 'C': 1, 'D': -1}
        charge_sum = balancecheck.reaction_charge(reaction, compound_charge)
        self.assertEqual(charge_sum, 0)

    def test_reaction_charge_non_zero_sum(self):
        reaction = parse_reaction('A[e] + (6) B[c] <=> (6) C[e] + (6) D[c]')
        compound_charge = {'A': 1, 'B': -1, 'C': 1, 'D': -1}
        charge_sum = balancecheck.reaction_charge(reaction, compound_charge)
        self.assertEqual(charge_sum, 5)

    def test_reaction_charge_nan_sum(self):
        reaction = parse_reaction('A[e] + (6) B[c] <=> (6) C[e] + (6) D[c]')
        compound_charge = {'A': 1, 'B': -1, 'C': 1}
        charge_sum = balancecheck.reaction_charge(reaction, compound_charge)
        self.assertTrue(math.isnan(charge_sum))

    def test_reaction_formula_normal_return(self):
        reaction = parse_reaction('A[e] + (6) B[c] <=> (6) C[e] + (6) D[c]')
        compound_formula = {
            'A': Formula.parse('C6H12O6'),
            'B': Formula.parse('O2'),
            'C': Formula.parse('CO2'),
            'D': Formula.parse('H2O')
        }
        left_form, right_form = balancecheck.reaction_formula(
            reaction, compound_formula)
        self.assertEqual(left_form, Formula.parse('C6H12O18'))
        self.assertEqual(right_form, Formula.parse('C6H12O18'))

    def test_reaction_formula_none_return(self):
        reaction = parse_reaction('A[e] + (6) B[c] <=> (6) C[e] + (6) D[c]')
        compound_formula = {
            'A': Formula.parse('C6H12O6'),
            'B': Formula.parse('O2'),
            'C': Formula.parse('CO2'),
        }
        result = balancecheck.reaction_formula(reaction, compound_formula)
        self.assertIsNone(result)


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2019-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

"""Implementation of Flux Balance Analysis."""

from __future__ import unicode_literals

import logging
import csv
from os import mkdir, path
from ..lpsolver import lp
import argparse
from six import iteritems
from ..expression import boolean
from ..lpsolver.lp import Expression, ObjectiveSense
from ..fluxanalysis import FluxBalanceProblem
from ..command import (ObjectiveMixin, SolverCommandMixin,
                       MetabolicMixin, Command)
from psamm.importer import write_yaml_model
from ..util import MaybeRelative

logger = logging.getLogger(__name__)


class GimmeCommand(MetabolicMixin, ObjectiveMixin,
                   SolverCommandMixin, Command):
    """Subset a metabolic model based on gene expression data."""

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--biomass-threshold', help='Threshold of objective reaction '
                                        'flux. Can be an absolute flux value '
                                        'or percentage with percent sign.',
            type=MaybeRelative, default=MaybeRelative('100%'))
        parser.add_argument(
            '--expression-threshold', type=float,
            help='Threshold for gene expression')
        parser.add_argument(
            '--transcriptome-file', type=argparse.FileType('r'),
            help='Two column file of gene ID to expression')
        parser.add_argument('--detail', action='store_true',
                            help='Show model statistics.')
        parser.add_argument(
            '--export-model', type=str, default=None,
            help='Path to directory for full model export.')
        super(GimmeCommand, cls).init_parser(parser)

    def run(self):
        solver = self._get_solver()
        model = self._model
        mm = model.create_metabolic_model()

        if self._args.export_model is not None:
            if path.exists('{}'.format(self._args.export_model)):
                logger.warning('Output directory {} already exists.'.format(
                    self._args.export_model))
                quit()

        base_gene_dict = {}
        for rxn in model.reactions:
            base_gene_dict[rxn.id] = rxn.genes

        threshold_value = parse_transcriptome_file(
            self._args.transcriptome_file,
            self._args.expression_threshold)

        exchange_exclude = [rxn for rxn in mm.reactions
                            if mm.is_exchange(rxn)]
        exchange_exclude.append(self._get_objective())
        mm_irreversible, reversible_gene_assoc, split_rxns, _ = \
            mm.make_irreversible(base_gene_dict,
                                 exclude_list=exchange_exclude)

        p = FluxBalanceProblem(mm_irreversible, solver)

        final_model, used_exchange, below_threshold_ids, \
            incon_score = \
            solve_gimme_problem(p, mm_irreversible,
                                self._get_objective(),
                                reversible_gene_assoc,
                                split_rxns, threshold_value,
                                self._args.biomass_threshold)

        print('# internal Reactions')
        for reaction in sorted(final_model):
            print(reaction)
        print('# Exchange Reactions')
        for reaction in used_exchange:
            print(reaction)
        used_below = str(len([x for x in
                              below_threshold_ids if x in final_model]))
        below = str(len(below_threshold_ids))
        if self._args.detail:
            logger.info('Used {} below threshold reactions out of {} total '
                        'below threshold reactions'.format(used_below, below))
            logger.info('Inconsistency Score: {}'.format(incon_score))
        if self._args.export_model:
            mkdir('{}'.format(self._args.export_model))
            reactions_to_discard = []
            for reaction in self._model.reactions:
                if reaction.id not in final_model:
                    reactions_to_discard.append(reaction.id)
            for rid in reactions_to_discard:
                self._model.reactions.discard(rid)
            compound_set = set()
            for reaction in self._model.reactions:
                for compound in reaction.equation.compounds:
                    compound_set.add(compound[0].name)
            compounds_to_discard = []
            for compound in self._model.compounds:
                if compound.id not in compound_set:
                    compounds_to_discard.append(compound.id)
            for cid in compounds_to_discard:
                self._model.compounds.discard(cid)
            exchanges_to_discard = []
            for key in self._model.exchange:
                if key.name not in compound_set:
                    exchanges_to_discard.append(key)
            for key in exchanges_to_discard:
                del self._model.exchange[key]
            limits_to_discard = []
            for key in self._model.limits:
                if key not in final_model:
                    limits_to_discard.append(key)
            for key in limits_to_discard:
                del self._model.limits[key]
            write_yaml_model(self._model, dest=self._args.export_model,
                             split_subsystem=False)


def solve_gimme_problem(problem, mm, biomass, reversible_gene_assoc,
                        split_rxns, transcript_values, threshold):
    """Formulates and Solves a GIMME model.

    Implementation of the GIMME algorithm (Becker and Pallson 2008).
    Accepts an irreversible metabolic model, LP problem, and GIMME specific
    data. Will add in relevant GIMME constraints to the LP problem and
    generates a contextualized model based on the media constraints and the
    transcriptome data provided.

    Args:
        problem: :class:`FluxBalanceProblem` to solve.
        mm: An irreversible metabolic model.
        biomass: Biomass reaction ID.
        reversible_gene_assoc: A dictionary of gene IDs from make_irreversible.
        split_rxns: A set of tuples of reaction IDs from make_irreversible.
        transcript_values: A dictionary returned from parse_transcriptome_file.
        threshold: A threshold that the biomass flux needs to stay above.
    """
    ci_dict = {}
    for reaction in mm.reactions:
        gene_string = reversible_gene_assoc.get(reaction)
        if gene_string is None:
            continue
        else:
            e = boolean.Expression(gene_string)
            ci = get_rxn_value(e._root, transcript_values)
            if ci is not None:
                ci_dict[reaction] = ci

    gimme_objective = Expression()

    threshold = threshold
    problem.maximize(biomass)
    if threshold.relative:
        threshold.reference = problem.get_flux(biomass)

    logger.info('Setting objective threshold to {}'.format(
        threshold))

    if problem.get_flux(biomass) < float(threshold):
        logger.warning('Input threshold '
                       'greater than maximum '
                       'biomass: {}'.format(problem.get_flux(biomass)))
        quit()
    problem.prob.add_linear_constraints(
        problem.get_flux_var(biomass) >= float(threshold))

    for key, value in iteritems(ci_dict):
        gimme_objective += problem.get_flux_var(key) * value
    problem.prob.define('gimme_objective', types=lp.VariableType.Continuous)
    obj = problem.prob.var('gimme_objective')
    problem.prob.add_linear_constraints(obj == gimme_objective)
    problem.prob.set_objective(obj)
    problem.prob.set_objective_sense(ObjectiveSense.Minimize)
    problem.prob.solve()
    used_rxns = set()
    sum_fluxes = 0
    for reaction in mm.reactions:
        if abs(problem.get_flux(reaction)) > 1e-12:
            original_id = reaction.replace('_forward', '')
            original_id = original_id.replace('_reverse', '')
            used_rxns.add(original_id)
            sum_fluxes += abs(problem.get_flux(reaction))
    used_below = 0
    used_above = 0
    off_below = 0
    off_above = 0
    final_model = set()
    original_reaction_set = set()
    for reaction in mm.reactions:
        if mm.is_exchange(reaction):
            continue
        else:
            reaction = reaction.replace('_forward', '')
            reaction = reaction.replace('_reverse', '')
            original_reaction_set.add(reaction)
    below_threshold_ids = set()
    for reaction in ci_dict.keys():
        reaction = reaction.replace('_forward', '')
        reaction = reaction.replace('_reverse', '')
        below_threshold_ids.add(reaction)

    used_below_list = []
    for reaction in original_reaction_set:
        if reaction in used_rxns:
            if reaction in below_threshold_ids:
                used_below += 1
                used_below_list.append(reaction)
                final_model.add(reaction)
            else:
                used_above += 1
                final_model.add(reaction)
        else:
            if reaction in below_threshold_ids:
                off_below += 1
            else:
                off_above += 1
                final_model.add(reaction)
    used_exchange = used_rxns - original_reaction_set

    return final_model, used_exchange, below_threshold_ids, \
        problem.prob.result.get_value(obj)


def parse_transcriptome_file(f, threshold):
    """Parses a file containing a gene to expression mapping.

    Parses a tab separated two column table. The first column contains gene
    IDs while the second column contains expression values. Will compare the
    expression values to the given threshold and return a dict with any
    genes that fall under the expression threshold, and the amount that they
    fall under the threshold.

    Args:
        f: a file containing a two column gene to expression table.
        threshold: Expression threshold value.
    """
    threshold_value = {}
    for row in csv.reader(f, delimiter=str('\t')):
        try:
            gene = row[0]
            gene = gene.replace('"', '')
            if float(row[1]) < threshold:
                threshold_value[gene] = threshold - float(row[1])
            elif float(row[1]) >= threshold:
                continue
        except ValueError:
            logger.warning('Invalid expression value '
                           'provided: gene: {} '
                           'value: {}'.format(row[0], row[1]))
    return threshold_value


def get_rxn_value(root, gene_dict):
    """Gets overall expression value for a reaction gene association.

    Recursive function designed to parse a gene expression and return
    a penalty value to use in the GIMME algorithm. This function is
    designed to return the value directly if the expression only has
    one gene. If the expression has multiple genes related by 'OR'
    associations, then it will return the highest lowest penalty value.
    If the genes are associated with 'AND' logic then the function
    will return the highest penalty value of the set of values.

    Args:
        root: object of boolean.Expression()._root
        gene_dict: dict of gene expression from parse_transcriptome_file
    """
    if type(root) == boolean.Variable:
        return gene_dict.get(root.symbol)
    elif type(root) is boolean.And:
        val_list = [x for x in
                    [get_rxn_value(i, gene_dict) for i in root._terms]
                    if x is not None]
        if len(val_list) == 0:
            return None
        else:
            return max(x for x in val_list if x is not None)
    elif type(root) is boolean.Or:
        val_list = [x for x in
                    [get_rxn_value(i, gene_dict)
                     for i in root._terms]]
        if None in val_list:
            return None
        else:
            return min(x for x in val_list if x is not None)



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

"""Representations of affine expressions and variables.

These classes can be used to represent affine expressions
and do manipulation and evaluation with substitutions of
particular variables.
"""

from __future__ import unicode_literals

import re
import numbers
import functools
from collections import Counter, defaultdict

import six
from six import string_types, iteritems, iterkeys, itervalues


@six.python_2_unicode_compatible
@functools.total_ordering
class Variable(object):
    """Represents a variable in an expression

    Equality of variables is based on the symbol.
    """

    def __init__(self, symbol):
        """Create variable with given symbol

        Symbol must start with a letter or underscore but
        can contain numbers in other positions.

        >>> Variable('x')
        Variable('x')
        """

        if not re.match(r'^[^\d\W]\w*\Z', symbol):
            raise ValueError('Invalid symbol {}'.format(symbol))
        self._symbol = str(symbol)

    @property
    def symbol(self):
        """Symbol of variable

        >>> Variable('x').symbol
        'x'
        """
        return self._symbol

    def simplify(self):
        """Return simplified expression

        The simplified form of a variable is always the
        variable itself.

        >>> Variable('x').simplify()
        Variable('x')
        """
        return self

    def substitute(self, mapping):
        """Return expression with variables substituted

        >>> Variable('x').substitute(lambda v: {'x': 567}.get(v.symbol, v))
        567
        >>> Variable('x').substitute(lambda v: {'y': 42}.get(v.symbol, v))
        Variable('x')
        >>> Variable('x').substitute(
        ...     lambda v: {'x': 123, 'y': 56}.get(v.symbol, v))
        123
        """

        return mapping(self)

    def __add__(self, other):
        return Expression({self: 1}) + other

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        return Expression({self: 1}) - other

    def __rsub__(self, other):
        return -self + other

    def __mul__(self, other):
        return Expression({self: 1}) * other

    def __rmul__(self, other):
        return self * other

    def __div__(self, other):
        return Expression({self: 1}) / other

    __truediv__ = __div__

    def __floordiv__(self, other):
        return Expression({self: 1}) // other

    def __neg__(self):
        return Expression({self: -1})

    def __eq__(self, other):
        """Check equality of variables"""
        if isinstance(other, Expression):
            return other == self
        return isinstance(other, Variable) and self._symbol == other._symbol

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        if isinstance(other, Variable):
            return self._symbol < other._symbol
        return NotImplemented

    def __hash__(self):
        return hash('Variable') ^ hash(self._symbol)

    def __str__(self):
        return self._symbol

    def __repr__(self):
        return str('Variable({})').format(repr(self._symbol))


@six.python_2_unicode_compatible
class Expression(object):
    """Represents an affine expression (e.g. 2x + 3y - z + 5)"""

    def __init__(self, *args):
        """Create new expression

        >>> Expression({ Variable('x'): 2 }, 3)
        Expression('2x + 3')
        >>> Expression({ Variable('x'): 1, Variable('y'): 1 })
        Expression('x + y')
        """

        if len(args) == 1 and isinstance(args[0], string_types):
            # Parse as string
            self._variables, self._offset = self._parse_string(args[0])
        elif len(args) <= 2:
            self._variables = {}
            self._offset = args[1] if len(args) >= 2 else 0

            variables = args[0] if len(args) >= 1 else {}
            for var, value in iteritems(variables):
                if not isinstance(var, Variable):
                    raise ValueError('Not a variable: {}'.format(var))
                if value != 0:
                    self._variables[var] = value
        else:
            raise TypeError('Expression() requires one or two arguments')

    def _parse_string(self, s):
        """Parse expression string

        Variables must be valid variable symbols and
        coefficients must be integers.
        """

        scanner = re.compile(r'''
            (\s+) |         # whitespace
            ([^\d\W]\w*) |  # variable
            (\d+) |         # number
            ([+-]) |        # sign
            (\Z) |          # end
            (.)             # error
        ''', re.DOTALL | re.VERBOSE)

        variables = defaultdict(int)
        offset = 0

        # Parse using four states:
        # 0: expect sign, variable, number or end (start state)
        # 1: expect sign or end
        # 2: expect variable or number
        # 3: expect sign, variable or end
        # All whitespace is ignored
        state = 0
        state_number = 1
        for match in re.finditer(scanner, s):
            whitespace, variable, number, sign, end, error = match.groups()
            if error is not None:
                raise ValueError(
                    'Invalid token in expression string: {}'.format(
                        match.group(0)))
            elif whitespace is not None:
                continue
            elif variable is not None and state in (0, 2, 3):
                variables[Variable(variable)] += state_number
                state = 1
            elif sign is not None and state in (0, 1, 3):
                if state == 3:
                    offset += state_number
                state_number = 1 if sign == '+' else -1
                state = 2
            elif number is not None and state in (0, 2):
                state_number = state_number * int(number)
                state = 3
            elif end is not None and state in (0, 1, 3):
                if state == 3:
                    offset += state_number
            else:
                raise ValueError(
                    'Invalid token in expression string: {}'.format(
                        match.group(0)))

        # Remove zero-coefficient elements
        variables = {var: value for var, value in iteritems(variables)
                     if value != 0}
        return variables, offset

    def simplify(self):
        """Return simplified expression.

        If the expression is of the form 'x', the variable will be returned,
        and if the expression contains no variables, the offset will be
        returned as a number.
        """
        result = self.__class__(self._variables, self._offset)
        if len(result._variables) == 0:
            return result._offset
        elif len(result._variables) == 1 and result._offset == 0:
            var, value = next(iteritems(result._variables))
            if value == 1:
                return var
        return result

    def substitute(self, mapping):
        """Return expression with variables substituted

        >>> Expression('x + 2y').substitute(
        ...     lambda v: {'y': -3}.get(v.symbol, v))
        Expression('x - 6')
        >>> Expression('x + 2y').substitute(
        ...     lambda v: {'y': Variable('z')}.get(v.symbol, v))
        Expression('x + 2z')
        """
        expr = self.__class__()
        for var, value in iteritems(self._variables):
            expr += value * var.substitute(mapping)
        return (expr + self._offset).simplify()

    def variables(self):
        """Return iterator of variables in expression"""
        return iter(self._variables)

    def __add__(self, other):
        """Add expressions, variables or numbers"""

        if isinstance(other, numbers.Number):
            return self.__class__(self._variables, self._offset + other)
        elif isinstance(other, Variable):
            return self + Expression({other: 1})
        elif isinstance(other, Expression):
            variables = Counter(self._variables)
            variables.update(other._variables)
            variables = {var: value for var, value in iteritems(variables)
                         if value != 0}
            return self.__class__(variables, self._offset + other._offset)
        return NotImplemented

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        """Subtract expressions, variables or numbers"""
        return self + -other

    def __rsub__(self, other):
        return -self + other

    def __mul__(self, other):
        """Multiply by scalar"""
        if isinstance(other, numbers.Number):
            if other == 0:
                return self.__class__()
            return self.__class__({var: value * other for var, value
                                   in iteritems(self._variables)},
                                  self._offset * other)
        return NotImplemented

    def __rmul__(self, other):
        return self * other

    def __div__(self, other):
        """Divide by scalar"""
        if isinstance(other, numbers.Number):
            return self.__class__({var: value / other for var, value
                                   in iteritems(self._variables)},
                                  self._offset / other)
        return NotImplemented

    __truediv__ = __div__

    def __floordiv__(self, other):
        if isinstance(other, numbers.Number):
            return self.__class__({var: value // other for var, value
                                   in iteritems(self._variables)},
                                  self._offset // other)
        return NotImplemented

    def __neg__(self):
        return self * -1

    def __eq__(self, other):
        """Expression equality"""
        if isinstance(other, Expression):
            return (self._variables == other._variables and
                    self._offset == other._offset)
        elif isinstance(other, Variable):
            # Check that there is just one variable in the expression
            # with a coefficient of one.
            return (self._offset == 0 and len(self._variables) == 1 and
                    next(iterkeys(self._variables)) == other and
                    next(itervalues(self._variables)) == 1)
        elif isinstance(other, numbers.Number):
            return len(self._variables) == 0 and self._offset == other
        return False

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        def all_terms():
            count_vars = 0
            for symbol, value in sorted((var.symbol, value) for var, value
                                        in iteritems(self._variables)):
                if value != 0:
                    count_vars += 1
                    yield symbol, value
            if self._offset != 0 or count_vars == 0:
                yield None, self._offset

        terms = []
        for i, spec in enumerate(all_terms()):
            symbol, value = spec
            if i == 0:
                # First term is special
                if symbol is None:
                    terms.append('{}'.format(value))
                elif abs(value) == 1:
                    terms.append(symbol if value > 0 else '-'+symbol)
                else:
                    terms.append('{}{}'.format(value, symbol))
            else:
                prefix = '+' if value >= 0 else '-'
                if symbol is None:
                    terms.append('{} {}'.format(prefix, abs(value)))
                elif abs(value) == 1:
                    terms.append('{} {}'.format(prefix, symbol))
                else:
                    terms.append('{} {}{}'.format(prefix, abs(value), symbol))
        return ' '.join(terms)

    def __repr__(self):
        return str('Expression({})').format(repr(str(self)))


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

import unittest

from psamm.datasource import entry


class TestDictEntries(unittest.TestCase):
    def test_create_compound_dict_entry(self):
        props = {
            'id': 'test_compound',
            'name': 'Test compound',
            'formula': 'H2O',
            'charge': 1,
            'custom': 'ABC'
        }
        e = entry.DictCompoundEntry(props)
        self.assertIsInstance(e, entry.CompoundEntry)
        self.assertEqual(e.id, 'test_compound')
        self.assertEqual(e.name, 'Test compound')
        self.assertEqual(e.formula, 'H2O')
        self.assertEqual(e.charge, 1)
        self.assertEqual(e.properties.get('custom'), 'ABC')

    def test_create_compound_dict_entry_copy(self):
        props = {
            'id': 'test_compound',
            'name': 'Test compound'
        }
        e = entry.DictCompoundEntry(props)
        e2 = entry.DictCompoundEntry(e)
        self.assertIsInstance(e2, entry.CompoundEntry)
        self.assertEqual(e2.id, 'test_compound')
        self.assertEqual(e2.name, 'Test compound')

    def test_create_compound_dict_entry_without_id(self):
        props = {
            'name': 'Compound 1',
            'formula': 'CO2'
        }
        with self.assertRaises(ValueError):
            e = entry.DictCompoundEntry(props)

    def test_create_compound_dict_entry_with_id_override(self):
        props = {
            'id': 'old_id',
            'name': 'Compound 1',
            'formula': 'CO2'
        }
        e = entry.DictCompoundEntry(props, id='new_id')
        self.assertEqual(e.id, 'new_id')
        self.assertEqual(e.properties['id'], 'new_id')

    def test_use_compound_dict_entry_setters(self):
        e = entry.DictCompoundEntry({}, id='new_id')
        e.formula = 'CO2'
        e.name = 'Compound 1'
        e.charge = 5
        self.assertEqual(e.formula, 'CO2')
        self.assertEqual(e.properties['formula'], 'CO2')
        self.assertEqual(e.name, 'Compound 1')
        self.assertEqual(e.charge, 5)

    def test_create_reaction_entry(self):
        props = {
            'id': 'reaction_1',
            'name': 'Reaction 1'
        }
        e = entry.DictReactionEntry(props)
        self.assertIsInstance(e, entry.ReactionEntry)
        self.assertEqual(e.id, 'reaction_1')
        self.assertEqual(e.name, 'Reaction 1')

    def test_create_reaction_entry_from_compound_entry(self):
        e = entry.DictCompoundEntry({
            'id': 'compound_1',
            'name': 'Compound 1'
        })
        with self.assertRaises(ValueError):
            e2 = entry.DictReactionEntry(e)

    def test_use_reaction_dict_entry_setters(self):
        e = entry.DictReactionEntry({}, id='reaction_1')
        e.name = 'Reaction 1'
        e.equation = 'A => B'
        e.genes = 'gene_1 and gene_2'
        self.assertEqual(e.name, 'Reaction 1')
        self.assertEqual(e.equation, 'A => B')
        self.assertEqual(e.genes, 'gene_1 and gene_2')
        self.assertEqual(e.properties['genes'], 'gene_1 and gene_2')

    def test_create_compartment_entry(self):
        props = {
            'id': 'c',
            'name': 'Cytosol'
        }
        e = entry.DictCompartmentEntry(props)
        self.assertIsInstance(e, entry.CompartmentEntry)
        self.assertEqual(e.id, 'c')
        self.assertEqual(e.name, 'Cytosol')


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>
# copyright 2017  Jon Lund Steffensen <jon_steffensen@uri.edi>
# Copyright 2020 Christopher Powers <c-11060@my.uri.edu>
# Copyright 2020 Elysha Sameth <esameth1@my.uri.edu>

from __future__ import unicode_literals

import re
import json
import logging

from collections import defaultdict
from six import text_type, string_types, integer_types, itervalues

from ..command import Command
from ..expression import boolean

logger = logging.getLogger(__name__)


_JSON_TYPES = string_types + integer_types + (
    dict, list, tuple, float, bool, type(None))
_QUOTE_REGEXP = re.compile(r'[\t\n]')


def _encode_value(value):
    if value is None:
        return ''
    if not isinstance(value, _JSON_TYPES):
        value = text_type(value)
    if (isinstance(value, string_types) and not _QUOTE_REGEXP.search(value) and
            value != ''):
        return value
    return json.dumps(value)


class ExportTableCommand(Command):
    """Export parts of the model as tab separated tables.

    This command can export various parts of the model as a TSV table.
    The output is written to standard output. The type argument is used
    to determine which part of the model to export:

    - reactions: Export reactions and reaction metadata
    - compounds: Export compounds and compound metadata
    - exchange: Export the list of exchange compounds/reactions
    - limits: Export list of internal flux limits
    - metadata: Export general model metadata
    - genes: Export list of genes to reactions
    """

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            'export', metavar='export_type',
            choices=['reactions', 'compounds', 'medium', 'exchange', 'limits',
                     'metadata', 'genes'],
            help='Type of model data to export')

    def run(self):
        export_type = self._args.export
        if export_type == 'reactions':
            self._reaction_export()
        elif export_type == 'compounds':
            self._compound_export()
        elif export_type == 'exchange' or export_type == 'medium':
            self._exchange_export()
        elif export_type == 'limits':
            self._limits_export()
        elif export_type == 'metadata':
            self._metadata_export()
        elif export_type == 'genes':
            self._gene_export()

    def _reaction_export(self):
        property_set = set()
        for reaction in self._model.reactions:
            property_set.update(reaction.properties)

        property_list_sorted = sorted(
            property_set, key=lambda x: (x != 'id', x != 'equation', x))

        print('\t'.join(
            [text_type(x) for x in property_list_sorted] + ['in_model']))

        model_reactions = set(self._model.model)
        for reaction in self._model.reactions:
            line_content = [reaction.properties.get(property)
                            for property in property_list_sorted]
            in_model = reaction.id in model_reactions
            line_content.append(in_model)
            print('\t'.join(_encode_value(value) for value in line_content))

    def _compound_export(self):
        compound_set = set()
        for compound in self._model.compounds:
            compound_set.update(compound.properties)

        compound_list_sorted = sorted(
            compound_set, key=lambda x: (x != 'id', x != 'name', x))

        print('\t'.join(
            [text_type(x) for x in compound_list_sorted] + ['in_model']))

        metabolic_model = self._model.create_metabolic_model()
        model_compounds = set(x.name for x in metabolic_model.compounds)
        for compound in self._model.compounds:
            line_content = [compound.properties.get(property)
                            for property in compound_list_sorted]
            in_model = compound.id in model_compounds
            line_content.append(in_model)
            print('\t'.join(_encode_value(value) for value in line_content))

    def _exchange_export(self):
        print('{}\t{}\t{}\t{}'.format('Compound ID', 'Reaction ID',
                                      'Lower Limit', 'Upper Limit'))
        default_flux = self._model.default_flux_limit

        for compound, reaction, lower, upper in itervalues(
                self._model.exchange):
            if lower is None:
                lower = -1 * default_flux

            if upper is None:
                upper = default_flux

            print('\t'.join(_encode_value(value) for value in (
                compound, reaction, lower, upper)))

    def _limits_export(self):
        print('{}\t{}\t{}'.format(
            'Reaction ID', 'Lower Limits', 'Upper Limits'))
        for reaction, lower, upper in itervalues(self._model.limits):
            print('\t'.join(_encode_value(value) for value in (
                reaction, lower, upper)))

    def _metadata_export(self):
        print('Model Name\t{}'.format(
            _encode_value(self._model.name)))
        print('Biomass Reaction\t{}'.format(
            _encode_value(self._model.biomass_reaction)))
        print('Default Flux Limits\t{}'.format(
            _encode_value(self._model.default_flux_limit)))

        if self._model.version_string is not None:
            print('Model version\t{}'.format(
                _encode_value(self._model.version_string)))

    def _gene_export(self):
        gene_assoc = defaultdict(list)

        model_reactions = set(self._model.model)

        for reaction in self._model.reactions:
            assoc = None
            if reaction.genes is None:
                continue
            elif isinstance(reaction.genes, string_types):
                assoc = boolean.Expression(reaction.genes)
            else:
                variables = [boolean.Variable(g) for g in reaction.genes]
                assoc = boolean.Expression(boolean.And(*variables))

            for gene in assoc.variables:
                gene_assoc[gene].append(reaction.id)

        print('{}\t{}\t{}'.format(
            'gene_id', 'reaction_in_model', 'reaction_not_in_model'))

        for gene, reaction in gene_assoc.items():
            in_model = []
            not_in_model = []
            for r in reaction:
                if r in model_reactions:
                    in_model.append(r)
                else:
                    not_in_model.append(r)
            print('{}\t{}\t{}'.format(
                str(gene), '#'.join([_encode_value(rxn) for rxn in in_model]),
                '#'.join([_encode_value(rxn) for rxn in not_in_model])))
            # print('{}\t{}'.format(str(gene), (
            #     '#'.join([_encode_value(value) for value in reaction]))))



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>
# Copyright 2016  Chao Liu <lcddzyx@gmail.com>
from __future__ import unicode_literals

import logging

from ..command import (Command, MetabolicMixin, LoopRemovalMixin,
                       ObjectiveMixin, SolverCommandMixin)
from .. import fluxanalysis, util
from .. import randomsparse

logger = logging.getLogger(__name__)


class RandomSparseNetworkCommand(MetabolicMixin, LoopRemovalMixin,
                                 ObjectiveMixin, SolverCommandMixin, Command):
    """Find a random minimal network of model reactions.

    Given a reaction to optimize and a threshold, delete reactions randomly
    until the flux of the reaction to optimize falls under the threshold.
    Keep deleting reactions until no more reactions can be deleted. By default
    this uses standard FBA (not tFBA). Since the internal fluxes are irrelevant
    the FBA and tFBA are equivalent for this purpose.

    The threshold can be specified as an absolute flux (e.g. '1.23') or a
    relative flux of the full model flux (e.g. '40.5%').
    """

    _supported_loop_removal = ['none', 'tfba']

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            'threshold', help='Threshold of objective reaction '
                              'flux. Can be an absolute flux value '
                              '(0.25) or percentage of maximum '
                              'biomass',
            type=util.MaybeRelative)
        parser.add_argument(
            '--type', help='Type of deletion to perform',
            choices=['reactions', 'exchange', 'genes', 'rxgenes'], type=str,
            required=True)
        super(RandomSparseNetworkCommand, cls).init_parser(parser)

    def run(self):
        reaction = self._get_objective()
        if not self._mm.has_reaction(reaction):
            self.fail(
                'Specified reaction is not in model: {}'.format(reaction))

        loop_removal = self._get_loop_removal_option()
        enable_tfba = loop_removal == 'tfba'
        if not enable_tfba:
            solver = self._get_solver()
        else:
            solver = self._get_solver(integer=True)

        p = fluxanalysis.FluxBalanceProblem(self._mm, solver)

        if enable_tfba:
            p.add_thermodynamic()

        try:
            p.maximize(reaction)
        except fluxanalysis.FluxBalanceError as e:
            self.report_flux_balance_error(e)

        threshold = self._args.threshold
        if threshold.relative:
            threshold.reference = p.get_flux(reaction)

        flux_threshold = float(threshold)

        logger.info('Flux threshold for {} is {}'.format(
            reaction, flux_threshold))

        if self._args.type == 'exchange':
            strategy = randomsparse.ReactionDeletionStrategy(
                self._mm, randomsparse.get_exchange_reactions(self._mm))
            entity = 'reactions'
        elif self._args.type == 'reactions':
            strategy = randomsparse.ReactionDeletionStrategy(
                self._mm)
            entity = 'reactions'
        elif self._args.type == 'rxgenes':
            strategy = randomsparse.GeneDeletionStrategy(
                self._mm, randomsparse.get_gene_associations(self._model))
            entity = 'genes'
        else:
            strategy = randomsparse.GeneDeletionStrategy(
                self._mm, randomsparse.get_gene_associations(self._model))
            entity = 'genes'

        if self._args.type == 'rxgenes':
            essential, deleted = randomsparse.random_sparse_return_all(
                strategy, p, reaction, flux_threshold)
            for i in self._mm.reactions:
                if i in deleted:
                    print('{}\t{}'.format(i, 0))
                else:
                    print('{}\t{}'.format(i, 1))
        else:
            essential, deleted = randomsparse.random_sparse(
                strategy, p, reaction, flux_threshold)

            logger.info('Essential {}: {}/{}'.format(
                entity, len(essential), len(strategy.entities)))
            logger.info('Deleted {}: {}/{}'.format(
                entity, len(deleted), len(strategy.entities)))

            for entity_id in sorted(strategy.entities):
                value = 0 if entity_id in deleted else 1
                print('{}\t{}'.format(entity_id, value))



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2016-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>
# Copyright 2020-2020  Elysha Sameth <esameth1@my.uri.edu>

"""Representation of compound/reaction entries in models."""

from __future__ import unicode_literals

import abc
from collections import Mapping

from six import add_metaclass


@add_metaclass(abc.ABCMeta)
class ModelEntry(object):
    """Abstract model entry.

    Provdides a base class for model entries which are representations of
    any entity (such as compound, reaction or compartment) in a model. An
    entity has an ID, and may have a name and filemark. The ID is a unique
    string identified within a model. The name is a string identifier for
    human consumption. The filemark indicates where the entry originates from
    (e.g. file name and line number). Any additional properties for an
    entity exist in ``properties`` which is any dict-like object mapping
    from string keys to any value type. The ``name`` entry in the dictionary
    corresponds to the name. Entries can be mutable, where the
    properties can be modified, or immutable, where the properties cannot be
    modified or where modifications are ignored. The ID is always immutable.
    """
    @abc.abstractproperty
    def id(self):
        """Identifier of entry."""

    @property
    def name(self):
        """Name of entry (or None)."""
        return self.properties.get('name')

    @abc.abstractproperty
    def properties(self):
        """Properties of entry as a :class:`Mapping` subclass (e.g. dict).

        Note that the properties are not generally mutable but may be mutable
        for specific subclasses. If the ``id`` exists in this dictionary, it
        must never change the actual entry ID as obtained from the ``id``
        property, even if other properties are mutable.
        """

    @abc.abstractproperty
    def filemark(self):
        """Position of entry in the source file (or None)."""

    def __repr__(self):
        return str('<{} id={!r}>').format(self.__class__.__name__, self.id)


class CompoundEntry(ModelEntry):
    """Abstract compound entry.

    Entry subclass for representing compounds. This standardizes the properties
    ``formula`` and ``charge``.
    """
    @property
    def formula(self):
        """Chemical formula of compound."""
        return self.properties.get('formula')

    @property
    def charge(self):
        """Compound charge value."""
        return self.properties.get('charge')


class ReactionEntry(ModelEntry):
    """Abstract reaction entry.

    Entry subclass for representing compounds. This standardizes the properties
    ``equation`` and ``genes``.
    """
    @property
    def equation(self):
        """Reaction equation."""
        return self.properties.get('equation')

    @property
    def genes(self):
        """Gene association expression."""
        return self.properties.get('genes')


class CompartmentEntry(ModelEntry):
    """Abstract compartment entry.

    Entry subclass for representing compartments.
    """


class _BaseDictEntry(ModelEntry):
    """Base class for concrete entries based on dictionary.

    The properties are mutable for this subclass. If ``id`` is None, the
    value corresponding to the ``id`` key in the dictionary is used. If this is
    not defined, a :class:`ValueError` is raised.
    """
    def __init__(self, abstract_type, properties={}, filemark=None, id=None):
        if isinstance(properties, abstract_type):
            self._id = id
            if self._id is None:
                self._id = properties.id
            self._properties = dict(properties.properties)
            if filemark is None:
                filemark = properties.filemark
        elif isinstance(properties, Mapping):
            self._id = id
            if self._id is None:
                if 'id' not in properties:
                    raise ValueError('id not defined in properties')
                self._id = properties['id']
            self._properties = dict(properties)
        else:
            raise ValueError('Invalid type of properties object')

        self._properties['id'] = self._id
        self._filemark = filemark

    @property
    def id(self):
        return self._id

    @ModelEntry.name.setter
    def name(self, value):
        self._properties['name'] = value

    @property
    def properties(self):
        return self._properties

    @property
    def filemark(self):
        return self._filemark


class DictCompoundEntry(CompoundEntry, _BaseDictEntry):
    """Compound entry backed by dictionary.

    The given properties dictionary must contain a key ``id`` with the
    identifier.

    Args:
        properties: dict or :class:`CompoundEntry` to construct from.
        filemark: Where the entry was parsed from (optional)
    """
    def __init__(self, *args, **kwargs):
        super(DictCompoundEntry, self).__init__(CompoundEntry, *args, **kwargs)

    @CompoundEntry.formula.setter
    def formula(self, value):
        self._properties['formula'] = value

    @CompoundEntry.charge.setter
    def charge(self, value):
        self._properties['charge'] = value


class DictReactionEntry(ReactionEntry, _BaseDictEntry):
    """Reaction entry backed by dictionary.

    The given properties dictionary must contain a key ``id`` with the
    identifier.

    Args:
        properties: dict or :class:`ReactionEntry` to construct from.
        filemark: Where the entry was parsed from (optional)
    """
    def __init__(self, *args, **kwargs):
        super(DictReactionEntry, self).__init__(ReactionEntry, *args, **kwargs)

    @ReactionEntry.equation.setter
    def equation(self, value):
        self._properties['equation'] = value

    @ReactionEntry.genes.setter
    def genes(self, value):
        self._properties['genes'] = value


class DictCompartmentEntry(CompartmentEntry, _BaseDictEntry):
    """Compartment entry backed by dictionary.

    The given properties dictionary must contain a key ``id`` with the
    identifier.

    Args:
        properties: dict or :class:`CompartmentEntry` to construct from.
        filemark: Where the entry was parsed from (optional)
    """
    def __init__(self, *args, **kwargs):
        super(DictCompartmentEntry, self).__init__(
            CompartmentEntry, *args, **kwargs)



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

from __future__ import unicode_literals

import time
import logging
from itertools import product

from ..command import (Command, SolverCommandMixin, MetabolicMixin,
                       ObjectiveMixin, LoopRemovalMixin, ParallelTaskMixin)
from ..util import MaybeRelative
from .. import fluxanalysis

logger = logging.getLogger(__name__)


class FluxVariabilityCommand(MetabolicMixin, SolverCommandMixin,
                             LoopRemovalMixin, ObjectiveMixin,
                             ParallelTaskMixin, Command):
    """Run flux variablity analysis on the model."""

    _supported_loop_removal = ['none', 'tfba']

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument('--threshold',
                            help='Threshold of objective reaction '
                            'flux. Can be an absolute flux value '
                            '(0.25) or percentage of maximum '
                            'biomass.)',
                            type=MaybeRelative,
                            default=MaybeRelative('100%'))
        super(FluxVariabilityCommand, cls).init_parser(parser)

    def run(self):
        """Run flux variability command"""

        # Load compound information
        def compound_name(id):
            if id not in self._model.compounds:
                return id
            return self._model.compounds[id].properties.get('name', id)

        reaction = self._get_objective()
        if not self._mm.has_reaction(reaction):
            self.fail(
                'Specified reaction is not in model: {}'.format(reaction))

        loop_removal = self._get_loop_removal_option()
        enable_tfba = loop_removal == 'tfba'
        if enable_tfba:
            solver = self._get_solver(integer=True)
        else:
            solver = self._get_solver()

        start_time = time.time()

        try:
            fba_fluxes = dict(fluxanalysis.flux_balance(
                self._mm, reaction, tfba=False, solver=solver))
        except fluxanalysis.FluxBalanceError as e:
            self.report_flux_balance_error(e)

        threshold = self._args.threshold
        if threshold.relative:
            threshold.reference = fba_fluxes[reaction]

        logger.info('Setting objective threshold to {}'.format(
            threshold))

        handler_args = (
            self._mm, solver, enable_tfba, float(threshold), reaction)
        executor = self._create_executor(
            FVATaskHandler, handler_args, cpus_per_worker=2)

        def iter_results():
            results = {}
            with executor:
                for (reaction_id, direction), value in executor.imap_unordered(
                        product(self._mm.reactions, (1, -1)), 16):
                    if reaction_id not in results:
                        results[reaction_id] = value
                        continue

                    other_value = results[reaction_id]
                    if direction == -1:
                        bounds = value, other_value
                    else:
                        bounds = other_value, value

                    yield reaction_id, bounds

            executor.join()

        for reaction_id, (lower, upper) in iter_results():
            rx = self._mm.get_reaction(reaction_id)
            rxt = rx.translated_compounds(compound_name)
            print('{}\t{}\t{}\t{}'.format(reaction_id, lower, upper, rxt))

        logger.info('Solving took {:.2f} seconds'.format(
            time.time() - start_time))


class FVATaskHandler(object):
    def __init__(self, model, solver, enable_tfba, threshold, reaction):
        self._problem = fluxanalysis.FluxBalanceProblem(model, solver)
        if enable_tfba:
            self._problem.add_thermodynamic()

        self._problem.prob.add_linear_constraints(
            self._problem.get_flux_var(reaction) >= threshold)

    def handle_task(self, reaction_id, direction):
        return self._problem.flux_bound(reaction_id, direction)



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2018-2020  Ke Zhang <kzhang@my.uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>
# Copyright 2020-2020  Elysha Sameth <esameth1@my.uri.edu>

from __future__ import unicode_literals

from itertools import count
from six import iteritems, text_type
from collections import defaultdict, Counter

from . import findprimarypairs

import logging

from .formula import Formula, Atom, ParseError
from .reaction import Direction, Reaction


logger = logging.getLogger(__name__)


def _graphviz_prop_string(d):
    return ','.join('{}="{}"'.format(k, text_type(v)) for k, v
                    in sorted(iteritems(d)))


class Entity(object):
    """Base class for graph entities."""

    def __init__(self, props={}):
        self._props = dict(props)

    @property
    def props(self):
        return self._props

    def __repr__(self):
        return '<{} id={}>'.format(
            self.__class__.__name__, self._props.get('id'))


class Graph(Entity):
    """Graph entity representing a collection of nodes and edges."""

    def __init__(self, props={}):
        super(Graph, self).__init__(props)
        self._nodes = set()
        self._edges = {}
        self._node_edges = {}
        self._edge_count = 0
        self._default_node_props = {}
        self._default_edge_props = {}
        self._nodes_id = {}
        self._nodes_original_id = defaultdict(list)

    def add_node(self, node):
        """add node to a Graph entity.
        node: Node entity.
        """
        self._nodes.add(node)
        self._nodes_id[node.props['id']] = node
        if 'type' in node.props:
            self.set_original_id(node)
        else:
            self._nodes_original_id[node.props['id']].append(node)

    def set_original_id(self, node):
        if node.props['type'] == 'cpd':
            original_id_string = ','.join([c.id for c in node.props['entry']])
        else:
            if node.props['type'] == 'Ex_rxn':
                original_id_string = node.props['id']
            else:
                original_id_string = ','.join(
                    [r.id for r in node.props['entry']])
        self._nodes_original_id[original_id_string].append(node)

    def get_node(self, node_id):
        """get Node object.
        Args:
            node_id: text_type(compound object) or text_type(a string that
                contains single or multiple reaction IDs).
        """
        return self._nodes_id[node_id]

    def add_edge(self, edge):
        if edge.source not in self._nodes or edge.dest not in self._nodes:
            raise ValueError('Edge nodes not in graph')

        s_nodes = tuple([edge.source, edge.dest])
        self._edges.setdefault(s_nodes, set()).add(edge)
        self._edge_count += 1

        self._node_edges.setdefault(edge.source, {}).setdefault(
            edge.dest, set()).add(edge)
        self._node_edges.setdefault(edge.dest, {}).setdefault(
            edge.source, set()).add(edge)

    @property
    def nodes(self):
        return iter(self._nodes)

    @property
    def node_count(self):
        return len(self._nodes)

    @property
    def nodes_id_dict(self):
        return self._nodes_id

    @property
    def nodes_original_id_dict(self):
        return self._nodes_original_id

    @property
    def edges(self):
        for pair, edges in iteritems(self._edges):
            for edge in edges:
                yield edge

    @property
    def edge_count(self):
        return self._edge_count

    def edges_for(self, node):
        return iteritems(self._node_edges.get(node, {}))

    @property
    def default_node_props(self):
        return self._default_node_props

    @property
    def default_edge_props(self):
        return self._default_edge_props

    def write_graphviz(self, f, width, height):
        """ Write the nodes and edges information into a dot file.

        Print a given graph object to a graph file in the dot format.
        This graph can then be converted to an image file using the
        graphviz program.

        Args:
            self: Graph entity, including nodes and edges entities.
            f: An empty file.
            width: Width of final metabolic map.
            height: Height of final metabolic map.
        """
        f.write('digraph {\n')
        if width is not None and height is not None:
            f.write('size = "{}, {}"; ratio = fill;\n'.format(width, height))

        if len(self._default_node_props) > 0:
            f.write(' node[{}];\n'.format(
                    _graphviz_prop_string(self._default_node_props)))

        if len(self._default_edge_props) > 0:
            f.write(' edge[{}];\n'.format(
                    _graphviz_prop_string(self._default_edge_props)))

        for k, v in iteritems(self.props):
            f.write(' {}="{}";\n'.format(k, v))

        for node in sorted(self.nodes, key=lambda k: k.props['id']):
            f.write(' "{}"[{}]\n'.format(node.props['id'],
                                         _graphviz_prop_string(node.props)))

        for edge in sorted(self.edges, key=lambda k: (
                k.source.props['id'], k.dest.props['id'], k.props.get('dir'))):
            f.write(' "{}" -> "{}"[{}]\n'.format(
                    edge.source.props['id'], edge.dest.props['id'],
                    _graphviz_prop_string(edge.props)))

        f.write('}\n')

    def write_graphviz_compartmentalized(self, f, compartment_tree,
                                         extracellular, width, height):
        """Function to write compartmentalized version of dot file
        for graph.

        Print a given graph object to a graph file in the dot format.
        In this graph, reaction nodes will be separated into different
        areas in the visualization based on the defined cellular
        compartments in the GEM.

        Args:
            self: Graph entity.
            f: An empty file.
            compartment_tree: a defaultdict of set, each element
                represents a compartment and its adjacent compartments.
            extracellular: the extracellular compartment in the model
            width: Width of final metabolic map.
            height: Height of final metabolic map..
        """
        f.write('digraph {\n')
        if width is not None and height is not None:
            f.write('size="{},{}"; ratio = fill;\n'.format(width, height))

        if len(self._default_node_props) > 0:
            f.write(' node[{}];\n'.format(
                    _graphviz_prop_string(self._default_node_props)))

        if len(self._default_edge_props) > 0:
            f.write(' edge[{}];\n'.format(
                    _graphviz_prop_string(self._default_edge_props)))

        for k, v in iteritems(self.props):
            f.write(' {}="{}";\n'.format(k, v))

        next_id = count(0)
        node_dicts = defaultdict(list)

        for node in self.nodes:
            node_dicts[node.props['compartment']].append(node)

        def edit_labels(string):
            return string.replace('-', '_')

        def write_node_props(f, node_list):
            for node in node_list:
                if 'id' not in node.props:
                    node.props['id'] = 'n{}'.format(next(next_id))
                f.write('  "{}"[{}]\n'.format(
                        node.props['id'], _graphviz_prop_string(node.props)))

        def dfs_recursive(graph, vertex, node_dict,
                          extracellular, f, path=[]):
            path.append(vertex)
            if vertex == extracellular:
                f.write(''.join(
                        [' subgraph cluster_{} '.format(edit_labels(vertex)),
                         '{\n  style=solid;\n  color=black;\n  penwidth=4;\n  '
                         'fontsize=35;\n',
                         '  label = "Compartment: {}"\n'.format
                         (edit_labels(vertex))]))
                for x in sorted(node_dict[vertex],
                                key=lambda k: k.props['id']):
                    f.write(' "{}"[{}]\n'.format(
                            x.props['id'], _graphviz_prop_string(x.props)))
            elif vertex != extracellular:
                f.write(''.join(
                        [' subgraph cluster_{} '.format(edit_labels(vertex)),
                         '{\n  style=dashed;\n  color=black;\n  '
                         'penwidth=4;\n  fontsize=35;\n',
                         '  label = "Compartment: {}"\n'.format
                         (edit_labels(vertex))]))
                for x in sorted(node_dict[vertex],
                                key=lambda k: k.props['id']):
                    f.write(' "{}"[{}]\n'.format(
                            x.props['id'], _graphviz_prop_string(x.props)))
            for neighbor in graph[vertex]:
                if neighbor not in path:
                    path = dfs_recursive(graph, neighbor, node_dict, path, f)
            f.write('}')
            return path

        dfs_recursive(compartment_tree, extracellular, node_dicts,
                      extracellular, f)

        for edge in sorted(self.edges, key=lambda k: (k.source.props['id'],
                                                      k.dest.props['id'],
                                                      k.props.get('dir'))):
            f.write(' "{}" -> "{}"[{}]\n'.format(
                    edge.source.props['id'], edge.dest.props['id'],
                    _graphviz_prop_string(edge.props)))

        f.write('}\n')

    def write_nodes_tables(self, f):
        """write a table file (.tsv) that contains nodes information.

        Write all node information from a given graph object into a
        tab separated table. This table will include IDs, shapes,
        fill colors, and labels. This table can be used to import
        the graph information to other programs.

        Args:
            self: Graph entity.
            f: An empty file.
        """

        properties = set()
        for node in self.nodes:
            if 'label' not in node.props:
                node.props['label'] = node.props['id']
            properties.update(node.props)
        if 'id' in properties:
            properties.remove('id')
        if 'label' in properties:
            properties.remove('label')
        if 'original_id' in properties:
            properties.remove('original_id')
        if 'entry' in properties:
            properties.remove('entry')
        properties = ['id'] + sorted(properties) + ['label']
        f.write('\t'.join(properties) + '\n')

        for node in sorted(self.nodes, key=lambda k: k.props['id']):
            # a = '\t'.join(text_type(node.props.get(x))
            #               for x in properties if x != 'label')
            a = '\t'.join(text_type(node.props.get(x))
                          for x in properties if x not in ['label', 'entry'])
            b = node.props['label'].replace('\n', ',')
            f.write('{}\t{}\n'.format(a, b))

    def write_edges_tables(self, f):
        """ Write a tab separated table that contains edges information,
        including edge source, edge dest, and edge properties.

        Write all edge information from a given graph object into a
        tab separated table. This table will include IDs, source nodes
        and destination nodes. This table can be used to import
        the graph information to other programs.

        Args:
            self: Graph entity.
            f: An empty TSV file.
        """
        properties = set()
        for edge in self.edges:
            properties.update(edge.props)

        properties = sorted(properties)
        header = ['source', 'target'] + properties
        f.write('\t'.join(header) + '\n')
        for edge in sorted(self.edges, key=lambda k: (edge.source.props['id'],
                                                      edge.dest.props['id'],
                                                      edge.props.get('dir'))):
            f.write('{}\t{}\t{}\n'.format(
                    edge.source.props['id'], edge.dest.props['id'],
                    '\t'.join(
                        text_type(edge.props.get(x)) for x in properties)))


class Node(Entity):
    """Node entity represents a vertex in the graph."""
    __hash__ = Entity.__hash__

    def __init__(self, props={}):
        super(Node, self).__init__(props)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return other._props == self._props
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)


class Edge(Entity):
    """Edge entity represents a connection between nodes."""
    __hash__ = Entity.__hash__

    def __init__(self, source, dest, props={}):
        super(Edge, self).__init__(props)
        self.source = source
        self.dest = dest

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            a = self._props == other._props
            b = self.source == other.source
            c = self.dest == other.dest
            if all(i is True for i in [a, b, c]):
                return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)


def get_compound_dict(model):
    """ Parse through model compounds and return a formula dict.

    This function will parse the compound information
    in a given model and return a dictionary of compound IDs to
    compound formula objects.

    Args:
        model: <class 'psamm.datasource.native.NativeModel'>
    """
    compound_formula = {}
    for compound in model.compounds:
        if compound.formula is not None:
            try:
                f = Formula.parse(compound.formula).flattened()
                if not f.is_variable():
                    compound_formula[compound.id] = f
                else:
                    logger.warning(
                        'Skipping variable formula {}: {}'.format(
                            compound.id, compound.formula))
            except ParseError as e:
                msg = (
                    'Error parsing formula'
                    ' for compound {}:\n{}\n{}'.format(
                        compound.id, e, compound.formula))
                if e.indicator is not None:
                    msg += '\n{}'.format(e.indicator)
                logger.warning(msg)
    return compound_formula


def make_network_dict(nm, mm, subset=None, method='fpp', element=None,
                      excluded_reactions=[], reaction_dict={}, analysis=None):
    """Create a dictionary of reactant/product pairs to reaction directions.

    Returns a dictionary that connects predicted reactant/product pairs
    to their original reaction directions. This dictionary is used when
    generating bipartite graph objects. This can be done either using
    the FindPrimaryPairs method to predict reactant/product pairs or
    by mapping all possible pairs.

    Args:
        nm: <class 'psamm.datasource.native.NativeModel'>, the native model
        mm: <class 'psamm.metabolicmodel.MetabolicModel'>, the metabolic model.
        subset: None or path to a file that contains a list of reactions or
            compound ids. By default, it is None. It defines which reaction
            need to be visualized.
        method: 'fpp' or 'no-fpp', the method used for visualization.
        element: Symbol of chemical atom, such as 'C' ('C' indicates carbon).
        excluded_reactions: a list that contains reactions excluded from
            visualization.
        reaction_dict: dictionary of FBA or FVA results. By default it is an
            empty dictionary.
        analysis: "None" type or a string indicates if FBA or FVA file is
            given in command line.
    """
    compound_formula = get_compound_dict(nm)
    if not compound_formula and (method == 'fpp' or element):
        logger.error(
            'Compound formulas are required for fpp or specific element '
            'visualizations, try --element all to visualize all pathways '
            'without compound formula input.')
        exit(1)

    if subset is not None:
        testing_list_raw = []
        for rxn in mm.reactions:
            if rxn in nm.reactions or mm.is_exchange(rxn):
                if rxn in subset:
                    if rxn not in excluded_reactions:
                        testing_list_raw.append(rxn)
                    else:
                        logger.warning(
                            'Reaction {} is in the subset and exclude file. '
                            'Reaction will be excluded.'.format(rxn))
    else:
        testing_list_raw = [rxn for rxn in mm.reactions
                            if (rxn in nm.reactions or mm.is_exchange(rxn)) and
                            rxn not in excluded_reactions]

    reaction_data = {}
    style_flux_dict = {}

    for rxn in testing_list_raw:
        flux = 0
        style = 'solid' if analysis is None else 'dotted'
        # direction = nm.reactions[rxn].equation.direction
        direction = mm.get_reaction(rxn).direction
        if rxn in reaction_dict:
            style = 'solid'
            reaction = mm.get_reaction(rxn)
            lower = reaction_dict[rxn][0]
            upper = reaction_dict[rxn][1]
            flux = lower * upper

            if flux == 0:
                if analysis == 'fba' or (lower == 0 and upper == 0):
                    style = 'dotted'
                elif lower == 0:
                    direction = Direction.Forward
                else:
                    direction = Direction.Reverse
            elif flux > 0:
                direction = Direction.Forward if (
                    analysis == 'fba' or lower > 0) else Direction.Reverse
            else:
                direction = Direction.Reverse if analysis == 'fba' \
                    else Direction.Both

            r = Reaction(direction, reaction.left, reaction.right)
            r_id = rxn
            mm.remove_reaction(rxn)
            mm.database.set_reaction(r_id, r)
            mm.add_reaction(r_id)
            if rxn in nm.reactions:
                nm.reactions[rxn].equation = r
        if rxn in nm.reactions:
            reaction_data[rxn] = (nm.reactions[rxn], direction)
        style_flux_dict[rxn] = (style, abs(flux))

    flux_list = sorted([f for s, f in style_flux_dict.values()])
    median = 1
    flux_list = list(filter(lambda x: x != 0, flux_list))

    testing_list = [rxn for rxn in testing_list_raw if not mm.is_exchange(rxn)]

    if flux_list:
        mid = len(flux_list) // 2
        median = flux_list[mid] if len(flux_list) % 2 else float(
            flux_list[mid-1] + flux_list[mid]) / 2.0
        if median < 1:
            median = 1

    for rxn, (style, flux) in style_flux_dict.items():
        new_flux = float(5 * flux) / float(median)
        new_flux = max(min(10, new_flux), 1)
        style_flux_dict.update({rxn: (style, new_flux)})

    full_pairs_dict = {}
    if method == 'fpp':
        element_weight = findprimarypairs.element_weight
        testing_list_update = []
        for r in testing_list:
            if all(cpd.name in compound_formula
                   for cpd, _ in nm.reactions[r].equation.compounds):
                testing_list_update.append(r)
            else:
                logger.warning(
                    'Reaction {} is excluded from visualization due to '
                    'missing or undefined compound formula'.format(r))
        reaction_pairs = [(r, nm.reactions[r].equation)
                          for r in testing_list_update]
        fpp_dict, _ = findprimarypairs.predict_compound_pairs_iterated(
            reaction_pairs, compound_formula, element_weight=element_weight)

        for rxn_id, fpp_pairs in iteritems(fpp_dict):
            compound_pairs = []
            for cpd_pair, transfer in iteritems(fpp_pairs[0]):
                if element is None:
                    compound_pairs.append(cpd_pair)
                else:
                    if any(Atom(element) in k for k in transfer):
                        compound_pairs.append(cpd_pair)
            (rxn_entry, rxn_dir) = reaction_data[rxn_id]
            full_pairs_dict[rxn_entry] = (sorted(compound_pairs), rxn_dir)

    elif method == 'no-fpp':
        for reaction in testing_list:
            compound_pairs = []
            for substrate in nm.reactions[reaction].equation.left:
                for product in nm.reactions[reaction].equation.right:
                    if element is None:
                        compound_pairs.append((substrate[0], product[0]))
                    else:
                        if Atom(element) in \
                                compound_formula[substrate[0].name] and \
                                Atom(element) in \
                                compound_formula[product[0].name]:
                            compound_pairs.append((substrate[0], product[0]))
            full_pairs_dict[nm.reactions[reaction]] = \
                (sorted(compound_pairs),
                 nm.reactions[reaction].equation.direction)

    return full_pairs_dict, style_flux_dict


def write_network_dict(network_dict):
    """Print out network dictionary object to a tab separated table.

    Print information from the dictionary "network_dict". This information
    includes four items: reaction ID, reactant, product, and direction.
    this can table can be used as an input to other graph visualization
    and analysis software.


     Args:
         network_dict: Dictionary object from make_network_dict()
     """
    for key, value in sorted(iteritems(network_dict), key=lambda x: str(x)):
        (cpair_list, dir) = value
        for (c1, c2) in cpair_list:
            print('{}\t{}\t{}\t{}'.format(key.id, c1, c2, dir_value(dir)))


def make_cpair_dict(filter_dict, args_method, args_combine, style_flux_dict,
                    hide_edges=[]):
    """Create a mapping from compound pair to a defaultdict containing
    lists of reactions for the forward, reverse, and both directions.

    Returns a dictionary that connects reactant/product pair to all reactions
    that contain this pair. Those reactions are stored in a dictionary and
    classified by reaction direction. For example:
    {(c1, c2): {'forward': [rxn1], 'back': [rxn3], 'both': [rxn4, rxn5]},
    (c3, c4): {...}, ...}

    Args:
        filter_dict: A dictionary mapping reaction entry to
            compound pairs (inside of the pairs there are cpd
            objects, not cpd IDs)
        args_method: a string, including 'fpp' and 'no-fpp'.
        args_combine: combine level, default = 0, optional: 1 and 2.
        style_flux_dict: a dictionary to set the edge style when fba or
            fva input is given.
        hide_edges: to determine edges between which compound pair need
            to be hidden.
    """

    new_id_mapping = {}
    new_style_flux_dict = {}
    rxn_count = Counter()
    cpair_dict = defaultdict(lambda: defaultdict(list))

    def make_mature_cpair_dict(cpair_dict, hide):
        new_cpair_dict = {}
        cpair_list = []
        for (c1, c2), rxns in sorted(iteritems(cpair_dict)):
            if (c1, c2) not in cpair_list and (text_type(c1),
                                               text_type(c2)) not in hide:
                new_rxns = rxns
                if (c2, c1) in cpair_dict:
                    if len(cpair_dict[(c2, c1)]['forward']) > 0:
                        for r in cpair_dict[(c2, c1)]['forward']:
                            new_rxns['back'].append(r)
                    if len(cpair_dict[(c2, c1)]['back']) > 0:
                        for r in cpair_dict[(c2, c1)]['back']:
                            new_rxns['forward'].append(r)
                    if len(cpair_dict[(c2, c1)]['both']) > 0:
                        for r in cpair_dict[(c2, c1)]['both']:
                            new_rxns['both'].append(r)
                    new_cpair_dict[(c1, c2)] = new_rxns
                    cpair_list.append((c1, c2))
                    cpair_list.append((c2, c1))
                else:
                    new_cpair_dict[(c1, c2)] = new_rxns
                    cpair_list.append((c1, c2))

        rxns_sorted_cpair_dict = defaultdict(lambda: defaultdict(list))
        for (c1, c2), rxns in sorted(iteritems(new_cpair_dict)):
            for direction, rlist in iteritems(rxns):
                rxns_sorted_cpair_dict[(c1, c2)][direction] = sorted(rlist)

        return rxns_sorted_cpair_dict

    if args_method != 'no-fpp':
        if args_combine == 0:
            for rxn, cpairs_dir in iteritems(filter_dict):
                have_visited = set()
                sub_pro = defaultdict(list)
                rxn_mixcpairs = defaultdict(list)
                for (c1, c2) in sorted(cpairs_dir[0]):
                    sub_pro[c1].append(c2)
                for k1, v1 in sorted(iteritems(sub_pro)):
                    if k1 not in have_visited:
                        rxn_count[rxn] += 1
                        have_visited.add(k1)
                        r_id = '{}_{}'.format(rxn.id, rxn_count[rxn])
                        new_id_mapping[r_id] = rxn
                        new_style_flux_dict[r_id] = style_flux_dict[rxn.id]
                        for v in v1:
                            rxn_mixcpairs[r_id].append((k1, v))
                        for k2, v2 in iteritems(sub_pro):
                            if k2 not in have_visited:
                                if k2 != k1:
                                    if v1 == v2:
                                        have_visited.add(k2)
                                        for vtest in v2:
                                            rxn_mixcpairs[r_id].append(
                                                (k2, vtest))
                for rxn_id, cpairs in iteritems(rxn_mixcpairs):
                    for (c1, c2) in cpairs:
                        if cpairs_dir[1] == Direction.Forward:
                            cpair_dict[(c1, c2)]['forward'].append(rxn_id)
                        elif cpairs_dir[1] == Direction.Reverse:
                            cpair_dict[(c1, c2)]['back'].append(rxn_id)
                        else:
                            cpair_dict[(c1, c2)]['both'].append(rxn_id)

        elif args_combine == 1 or args_combine == 2:
            for rxn, cpairs_dir in iteritems(filter_dict):
                cpd_rid = {}
                have_visited = set()
                for (c1, c2) in sorted(cpairs_dir[0]):
                    if c1 not in have_visited:
                        if c2 not in have_visited:
                            rxn_count[rxn] += 1
                            rxn_id = '{}_{}'.format(
                                rxn.id, rxn_count[rxn])
                            new_id_mapping[rxn_id] = rxn
                            new_style_flux_dict[rxn_id] = \
                                style_flux_dict[rxn.id]
                            have_visited.add(c1)
                            have_visited.add(c2)
                            cpd_rid[c1] = rxn_id
                            cpd_rid[c2] = rxn_id
                        else:
                            rxn_id = cpd_rid[c2]
                            have_visited.add(c1)
                            cpd_rid[c1] = rxn_id
                    else:
                        rxn_id = cpd_rid[c1]
                        have_visited.add(c2)
                        cpd_rid[c2] = rxn_id

                    if cpairs_dir[1] == Direction.Forward:
                        cpair_dict[(c1, c2)]['forward'].append(rxn_id)
                    elif cpairs_dir[1] == Direction.Reverse:
                        cpair_dict[(c1, c2)]['back'].append(rxn_id)
                    else:
                        cpair_dict[(c1, c2)]['both'].append(rxn_id)
    else:
        for rxn, cpairs_dir in iteritems(filter_dict):
            for (c1, c2) in cpairs_dir[0]:
                r_id = rxn.id
                new_id_mapping[r_id] = rxn
                new_style_flux_dict[r_id] = style_flux_dict[rxn.id]
                if cpairs_dir[1] == Direction.Forward:
                    cpair_dict[(c1, c2)]['forward'].append(r_id)
                elif cpairs_dir[1] == Direction.Reverse:
                    cpair_dict[(c1, c2)]['back'].append(r_id)
                else:
                    cpair_dict[(c1, c2)]['both'].append(r_id)

    rxns_sorted_cpair_dict = make_mature_cpair_dict(cpair_dict, hide_edges)
    return rxns_sorted_cpair_dict, new_id_mapping, new_style_flux_dict


def make_bipartite_graph_object(cpairs_dict, new_id_mapping, method,
                                args_combine, model_compound_entries,
                                new_style_flux_dict, analysis=None):
    """ Makes a bipartite graph object from a cpair_dict object.

    Start from empty graph() and cpair dict to make a graph object.
    Nodes only have rxn/cpd ID and rxn/cpd entry info in this initial
    graph. The information can be modified to add on other properties
    like color, or names.

    Args:
        cpairs_dict: defaultdict of compound_pair:
            defaultdict of direction: reaction list.
            e.g. {(c1, c2): {'forward":[rx1],
            'both':[rx2}}.
        new_id_mapping: dictionary of rxn_id_suffix: rxn_id.
        method: options=['fpp', 'no-fpp', file_path].
        args_combine: Command line argument, could be 0, 1, 2.
        model_compound_entries: dict of cpd_id:compound_entry.
        new_style_flux_dict: a dictionary to determine the edge style with
            the new reaction IDs.
    return: A graph object that contains basic nodes and edges.
        only ID and rxn/cpd entry are in node properties,
        no features like color, shape.
    """
    g = Graph()
    g._default_node_props['fontname'] = 'Arial'
    g._default_node_props['fontsize'] = 12

    def add_graph_nodes(g, cpairs_dict, method, new_id_mapping, args_combine,
                        model_compound_entries):
        """ Create compound and reaction nodes, adding them to
            empty graph object.

        Args:
            g: An empty Graph object.
            cpairs_dict: defaultdict of compound_pair:
                defaultdict of direction:
                reaction list. e.g. {(c1, c2): {'forward":[rx1],
                    'both':[rx2}}.
            method: Command line argument,
                options=['fpp', 'no-fpp', file_path].
            new_id_mapping: dictionary of rxn_id_suffix: rxn_id.
            args_combine: Command line argument,
                could be 0, 1, 2. By default it is 0.
            model_compound_entries: cpd id map to cpd entry of native model.
        return: A graph object that contains a set of nodes.
        """
        graph_nodes = set()
        for cpair, reactions in sorted(iteritems(cpairs_dict)):
            for c in cpair:
                if c not in graph_nodes:
                    node = Node({
                        'id': text_type(c),
                        'entry': [model_compound_entries[c.name]],
                        'compartment': c.compartment,
                        'type': 'cpd'})
                    g.add_node(node)
                    graph_nodes.add(c)
            for direction, rlist in iteritems(reactions):
                if method != 'no-fpp' and args_combine == 2:
                    real_rxns = [new_id_mapping[r] for r in rlist]
                    rxn_string = text_type(','.join(rlist))
                    if rxn_string not in graph_nodes:
                        rnode = Node({
                            'id': text_type(','.join(rlist)),
                            'entry': real_rxns,
                            'compartment': c.compartment,
                            'type': 'rxn'})
                        g.add_node(rnode)
                        graph_nodes.add(rxn_string)
                else:
                    for sub_rxn in rlist:
                        rxn_ob = new_id_mapping[sub_rxn]
                        if sub_rxn not in graph_nodes:
                            rnode = Node({
                                'id': text_type(sub_rxn),
                                'entry': [rxn_ob],
                                'compartment': c.compartment,
                                'type': 'rxn'})
                            g.add_node(rnode)
                            graph_nodes.add(sub_rxn)
        return g

    def add_edges(g, cpairs_dict, method, args_combine, new_style_flux_dict,
                  analysis):
        """ Add edges to the graph object obtained in last step.

        Args:
            g: A graph object contains a set of nodes.
            cpairs_dict: A defaultdict of compound_pair: defaultdict
                of direction: reaction list. e.g. {(c1, c2):
                {'forward":[rx1], 'both':[rx2}}.
            method: Command line argument, options=
                ['fpp', 'no-fpp', file_path].
            # split: Command line argument, True or False.
                By default split = False.
            args_combine: Command line argument, an
                integer(could be 0, 1 or 2).
            new_style_flux_dict: A dictsionary of reaction is maps to edge
                style and edge width.
            analysis: "None" type or a string indicates if FBA or FVA file is
                given in command line.
        """
        edge_list = []
        for (c1, c2), value in iteritems(cpairs_dict):
            for direction, rlist in iteritems(value):
                new_rlist = ','.join(rlist)
                if args_combine == 0 or args_combine == 1 or \
                        method == 'no-fpp':
                    for sub_rxn in rlist:
                        test1 = c1, sub_rxn
                        if test1 not in edge_list:
                            edge_list.append(test1)
                            g.add_edge(Edge(
                                g.get_node(text_type(c1)),
                                g.get_node(text_type(sub_rxn)),
                                {'dir': direction,
                                 'style': new_style_flux_dict[sub_rxn][0],
                                 'penwidth': new_style_flux_dict[sub_rxn][1]}))

                        test2 = c2, sub_rxn
                        if test2 not in edge_list:
                            edge_list.append(test2)
                            g.add_edge(Edge(
                                g.get_node(text_type(sub_rxn)),
                                g.get_node(text_type(c2)),
                                {'dir': direction,
                                 'style': new_style_flux_dict[sub_rxn][0],
                                 'penwidth': new_style_flux_dict[sub_rxn][1]}))

                else:
                    test1 = c1, new_rlist
                    test2 = c2, new_rlist
                    sub_rxn = list(new_rlist.split(','))

                    style_list = set(new_style_flux_dict[rxn][0] for rxn in
                                     sub_rxn)
                    style = style_list.pop() if len(set(style_list)) == 1 \
                        else 'solid'

                    flux = sum([new_style_flux_dict[rxn][1] for rxn in
                                sub_rxn], 0) if analysis else 0
                    flux = max(min(10, flux), 1)

                    if test1 not in edge_list:
                        edge_list.append(test1)
                        g.add_edge(Edge(
                            g.get_node(text_type(c1)),
                            g.get_node(text_type(new_rlist)),
                            {'dir': direction, 'style': style,
                             'penwidth': flux}))

                    if test2 not in edge_list:
                        edge_list.append(test2)
                        g.add_edge(Edge(
                            g.get_node(text_type(new_rlist)),
                            g.get_node(text_type(c2)),
                            {'dir': direction, 'style': style,
                             'penwidth': flux}))
        return g

    g = add_graph_nodes(g, cpairs_dict, method, new_id_mapping,
                        args_combine, model_compound_entries)
    g = add_edges(g, cpairs_dict, method, args_combine, new_style_flux_dict,
                  analysis)
    return g


def make_compound_graph(network_dictionary):
    g = Graph()
    compound_nodes = []
    edge_list = []
    for reaction, (cpair_list, dir) in iteritems(network_dictionary):
        for (c1, c2) in cpair_list:
            if c1 not in compound_nodes:
                g.add_node(Node({'id': text_type(c1), 'entry': c1}))
                compound_nodes.append(c1)
            if c2 not in compound_nodes:
                g.add_node(Node({'id': text_type(c2), 'entry': c2}))
                compound_nodes.append(c2)
            cpair_sorted = sorted([c1.name, c2.name])
            edge = Edge(g.get_node(text_type(c1)),
                        g.get_node(text_type(c2)),
                        props={'id': '{}_{}_{}'.format(
                            cpair_sorted[0], cpair_sorted[1],
                            dir_value(dir)), 'dir': dir_value(dir)})
            if edge.props['id'] not in edge_list:
                g.add_edge(edge)
                edge_list.append(edge.props['id'])
    return g


def dir_value(direction):
    """ Assign value to different reaction directions"""
    if direction == Direction.Forward:
        return 'forward'
    elif direction == Direction.Reverse:
        return 'back'
    else:
        return 'both'
